# Copyright 2024 CrackNuts. All rights reserved.

from cracknuts.trace.trace import ScarrTraceDataset
from cracknuts.trace.trace import NumpyTraceDataset

__all__ = ["ScarrTraceDataset", "NumpyTraceDataset"]



# Copyright 2024 CrackNuts. All rights reserved.

import struct

__version__ = "1.0.0"

DEFAULT_PORT = 9761

DEFAULT_OPERATOR_PORT = 9760

MAGIC_STR = b"CRAK"

# ┌───────────────────────────────────────────┬───────┐
# │--------------Request Header---------------│-------│
# │Magic │Version│Direction│Command│RFU│Length│PayLoad│
# │-------------------------------------------│-------│
# │ 4B   │  2B   │   1B    │   2B  │2B │  4B  │$Length│
# │-------------------------------------------│-------│
# │'CRAK'│   1   │   'S'   │       │   │      │       │
# └───────────────────────────────────────────┴───────┘

REQ_HEADER_FORMAT = ">4sH1sHHI"
REQ_HEADER_SIZE = struct.calcsize(REQ_HEADER_FORMAT)

# ┌──────────────────────────────────────┬───────┐
# │--------------Response Header---------│-------│
# │Magic │Version│Direction│Status│Length│PayLoad│
# │--------------------------------------│-------│
# │ 4B   │  2B   │   1B    │  2B  │  4B  │$Length│
# │--------------------------------------│-------│
# │'CRAK'│   1   │   'R'   │      │      │       │
# └──────────────────────────────────────┴───────┘

RES_HEADER_FORMAT = ">4sH1sHI"
RES_HEADER_SIZE = struct.calcsize(RES_HEADER_FORMAT)

DIRECTION_SEND = b"S"
DIRECTION_RESPONSE = b"R"

# Status Code
STATUS_OK = 0x0000
STATUS_BUSY = 0x0001
STATUS_ERROR = 0x8000
STATUS_COMMAND_UNSUPPORTED = 0x8001
STATUS_NUT_TIMEOUT = 0x8002
STATUS_ANALOG_CHANNEL_NONE_DATA = 0x8003
STATUS_DIGITAL_CHANNEL_NONE_DATA = 0x8004

STATUS_DESCRIPTION = {
    STATUS_OK: "OK",
    STATUS_BUSY: "BUSY",
    STATUS_ERROR: "ERROR",
    STATUS_COMMAND_UNSUPPORTED: "COMMAND UNSUPPORTED",
    STATUS_NUT_TIMEOUT: "NUT EXECUTE TIMEOUT",
    STATUS_ANALOG_CHANNEL_NONE_DATA: "ANALOG CHANNEL HAVE NO DATA",
    STATUS_DIGITAL_CHANNEL_NONE_DATA: "DIGITAL CHANNEL HAVE NO DATA",
}


def version():
    return __version__


class Command:
    """
    The generic command proxy configuration class defines the `CNP` communication code for general devices.
    For details, refer to the `CNP` protocol documentation.

    """

    GET_ID = 0x0001
    GET_NAME = 0x0002
    GET_VERSION = 0x0003

    CRACKER_READ_REGISTER = 0x0004
    CRACKER_WRITE_REGISTER = 0x0005

    SET_LOGGING_LEVEL = 0x0006

    OSC_ANALOG_CHANNEL_ENABLE = 0x0100
    OSC_ANALOG_COUPLING = 0x0101
    OSC_ANALOG_VOLTAGE = 0x0102
    OSC_ANALOG_BIAS_VOLTAGE = 0x0103
    OSC_ANALOG_GAIN = 0x0104
    OSC_ANALOG_GAIN_RAW = 0x0105
    OSC_CLOCK_BASE_FREQ_MUL_DIV = 0x0106
    OSC_CLOCK_SAMPLE_DIVISOR = 0x0107
    OSC_CLOCK_SAMPLE_PHASE = 0x0108
    OSC_CLOCK_NUT_DIVISOR = 0x0109
    OSC_CLOCK_UPDATE = 0x10A
    OSC_CLOCK_SIMPLE = 0x10B

    OSC_DIGITAL_CHANNEL_ENABLE = 0x0110
    OSC_DIGITAL_VOLTAGE = 0x0111

    OSC_TRIGGER_MODE = 0x0151

    OSC_ANALOG_TRIGGER_SOURCE = 0x0150
    OSC_DIGITAL_TRIGGER_SOURCE = 0x0122

    OSC_TRIGGER_EDGE = 0x0152
    OSC_TRIGGER_EDGE_LEVEL = 0x0153

    OSC_ANALOG_TRIGGER_VOLTAGE = 0x0123

    OSC_SAMPLE_DELAY = 0x0124

    OSC_SAMPLE_LENGTH = 0x0125
    OSC_SAMPLE_RATE = 0x0128

    OSC_SINGLE = 0x0126

    OSC_IS_TRIGGERED = 0x0127
    OSC_FORCE = 0x0129

    OSC_GET_ANALOG_WAVES = 0x0130
    OSC_GET_DIGITAL_WAVES = 0x0130

    NUT_ENABLE = 0x0200
    NUT_VOLTAGE = 0x0201
    NUT_VOLTAGE_RAW = 0x0203
    NUT_CLOCK_ENABLE = 0x0204
    NUT_CLOCK = 0x0202
    NUT_INTERFACE = 0x0210
    NUT_TIMEOUT = 0x0224

    CRACKER_SERIAL_BAUD = 0x0220
    CRACKER_SERIAL_WIDTH = 0x0221
    CRACKER_SERIAL_STOP = 0x0222
    CRACKER_SERIAL_ODD_EVE = 0x0223
    CRACKER_SERIAL_DATA = 0x022A

    CRACKER_SPI_CPOL = 0x0230
    CRACKER_SPI_CPHA = 0x0231
    CRACKER_SPI_DATA_LEN = 0x0232
    CRACKER_SPI_FREQ = 0x0233
    CRACKER_SPI_TIMEOUT = 0x0234
    CRACKER_SPI_DATA = 0x023A
    CRACKER_SPI_TRANSCEIVE = 0x023C
    CRACKER_SPI_ENABLE = 0x023D
    CRACKER_SPI_RESET = 0x023E
    CRACKER_SPI_CONFIG = 0x023F

    CRACKER_I2C_FREQ = 0x0240
    CRACKER_I2C_TIMEOUT = 0x0244
    CRACKER_I2C_TRANSCEIVE = 0x024A
    CRACKER_I2C_ENABLE = 0x024B
    CRACKER_I2C_RESET = 0x024C
    CRACKER_I2C_CONFIG = 0x024D

    CRACKER_UART_ENABLE = 0x0260
    CRACKER_UART_RESET = 0x0261
    CRACKER_UART_CONFIG = 0x0262
    CRACKER_UART_TRANSCEIVE = 0x0263
    CRACKER_UART_RECEIVE_FIFO_REMAINED = 0x0264
    CRACKER_UART_CRACKER_UART_RECEIVE_FIFO_DUMP = 0x0265
    CRACKER_UART_CRACKER_UART_RECEIVE_CLEAR = 0x0266

    CRACKER_CAN_FREQ = 0x0250
    CRACKER_CAN_TIMEOUT = 0x0254
    CRACKER_CA_DATA = 0x025A

    GET_BITSTREAM_VERSION = 0x0007


def _build_req_message(command: int, rfu: int = 0, payload: bytes | None = None):
    content = struct.pack(
        REQ_HEADER_FORMAT,
        MAGIC_STR,
        int(__version__.split(".")[0]),
        DIRECTION_SEND,
        command,
        rfu,
        0 if payload is None else len(payload),
    )
    if payload is not None:
        content += payload

    return content


def _build_res_message(status: int, payload: bytes | None = None):
    content = struct.pack(
        RES_HEADER_FORMAT,
        MAGIC_STR,
        int(__version__.split(".")[0]),
        DIRECTION_RESPONSE,
        status,
        0 if payload is None else len(payload),
    )
    if payload is not None:
        content += payload

    return content


def build_send_message(command: int, rfu: int = 0, payload: bytes | None = None):
    return _build_req_message(command, rfu, payload)


def build_response_message(status: int, payload: bytes | None = None):
    return _build_res_message(status, payload)



# Copyright 2024 CrackNuts. All rights reserved.

import re
import struct

from cracknuts.cracker import protocol, serial
from cracknuts.cracker.cracker_basic import ConfigBasic, CrackerBasic


class ConfigS1(ConfigBasic):
    def __init__(self):
        super().__init__()

        self.nut_enable = False
        self.nut_clock_enable = False
        self.nut_voltage = 3.5
        self.nut_clock = 8000
        self.nut_voltage_raw: int | None = None
        self.nut_interface: int | None = None
        self.nut_timeout: int | None = None

        self.cracker_uart_enable: bool | None = False
        self.cracker_uart_config: dict | None = {
            "baudrate": 115200,
            "bytesize": serial.Bytesize.EIGHTBITS,
            "parity": serial.Parity.PARITY_NONE,
            "stopbits": serial.Stopbits.STOPBITS_ONE,
        }

        self.cracker_spi_enable: bool | None = False
        self.cracker_spi_config: dict | None = {
            "speed": 10_000,
            "cpol": serial.SpiCpol.SPI_CPOL_LOW,
            "cpha": serial.SpiCpha.SPI_CPHA_LOW,
        }

        self.cracker_i2c_enable: bool | None = False
        self.cracker_i2c_config: dict | None = {
            "dev_addr": 0x00,
            "speed": serial.I2cSpeed.STANDARD_100K,
        }

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()


class CrackerS1(CrackerBasic[ConfigS1]):
    def get_default_config(self) -> ConfigS1:
        return ConfigS1()

    def sync_config_to_cracker(self):
        config = self.get_current_config()
        self._nut_set_enable(config.nut_enable)
        self.nut_voltage(config.nut_voltage)
        self._nut_set_clock_enable(config.nut_clock_enable)
        self.nut_clock_freq(config.nut_clock)
        for k, v in config.osc_analog_channel_enable.items():
            self._osc_set_analog_channel_enable(k, v)
            self.osc_analog_gain(k, config.osc_analog_gain.get(k, False))
        self.osc_sample_length(config.osc_sample_length)
        self.osc_sample_delay(config.osc_sample_delay)
        self.osc_sample_clock(config.osc_sample_clock)
        self.osc_sample_clock_phase(config.osc_sample_phase)
        self.osc_trigger_source(config.osc_analog_trigger_source)
        self.osc_trigger_mode(config.osc_trigger_mode)
        self.osc_trigger_edge(config.osc_analog_trigger_edge)
        self.osc_trigger_level(config.osc_analog_trigger_edge_level)

    def register_read(self, base_address: int, offset: int) -> tuple[int, bytes | None]:
        """
        Read register.

        :param base_address: Base address of the register.
        :type base_address: int
        :param offset: Offset of the register.
        :type offset: int
        :return: The device response status and the value read from the register or None if an exception is raised.
        :rtype: tuple[int, bytes | None]
        """
        payload = struct.pack(">II", base_address, offset)
        self._logger.debug(f"register_read payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_READ_REGISTER, payload=payload)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    def register_write(self, base_address: int, offset: int, data: bytes | int | str) -> tuple[int, None]:
        """
        Write register.

        :param base_address: Base address of the register.
        :type base_address: int
        :param offset: Offset of the register.
        :type offset: int
        :param data: Data to write.
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(data, str):
            if data.startswith("0x") or data.startswith("0X"):
                data = data[2:]
            data = bytes.fromhex(data)
        if isinstance(data, int):
            data = struct.pack(">I", data)
        payload = struct.pack(">II", base_address, offset) + data
        self._logger.debug(f"register_write payload: {payload.hex()}")
        status, _ = self.send_with_command(protocol.Command.CRACKER_WRITE_REGISTER, payload=payload)
        return status, None

    def osc_analog_enable(self, channel: int | str) -> tuple[int, None]:
        """
        Enable osc analog.

        :param channel: Channel to enable, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._osc_set_analog_channel_enable(channel, True)

    def osc_analog_disable(self, channel: int | str) -> tuple[int, None]:
        """
        Disable osc analog.

        :param channel: Channel to enable, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._osc_set_analog_channel_enable(channel, False)

    def _osc_set_analog_channel_enable(self, channel: int, enable: bool) -> tuple[int, None]:
        """
        Set analog channel enable.

        :param channel: Channel to enable, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :param enable: Enable or disable.
        :type enable: bool
        :return: The device response status
        :rtype: tuple[int, None]
        """
        channels = ("A", "B")
        if isinstance(channel, str):
            channel = channel.upper()
            if channel in channels:
                channel = channels.index(channel)
            else:
                self._logger.error("Channel error, it should in 'A' or 'B'")
                return self.NON_PROTOCOL_ERROR, None
        else:
            if channel > 1:
                self._logger.error("Channel error, it should in 0 or 1")
                return self.NON_PROTOCOL_ERROR, None
        final_enable = self._config.osc_analog_channel_enable | {channel: enable}
        mask = 0
        if final_enable.get(0):
            mask |= 1
        if final_enable.get(1):
            mask |= 1 << 1
        if final_enable.get(2):
            mask |= 1 << 2
        if final_enable.get(3):
            mask |= 1 << 3
        if final_enable.get(4):
            mask |= 1 << 4
        if final_enable.get(5):
            mask |= 1 << 5
        if final_enable.get(6):
            mask |= 1 << 6
        if final_enable.get(7):
            mask |= 1 << 7
        if final_enable.get(8):
            mask |= 1 << 8
        payload = struct.pack(">B", mask)
        self._logger.debug(f"Scrat analog_channel_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_ANALOG_CHANNEL_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_analog_channel_enable = final_enable
        return status, None

    def osc_trigger_mode(self, mode: int | str) -> tuple[int, None]:
        """
        Set trigger mode.

        :param mode: Trigger mode. Trigger mode can be one of ("EDGE", "PATTERN") or ("E", "P"),
                     or the index of the string value.
        :type mode: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        modes1 = ("EDGE", "PATTERN")
        modes2 = ("E", "P")
        if isinstance(mode, str):
            mode = mode.upper()
            if mode in modes1:
                mode = modes1.index(mode)
            elif mode in modes2:
                mode = modes2.index(mode)
            else:
                self._logger.error(f"Mode must be one of {modes1} or {modes2} if specified as a string.")
                return self.NON_PROTOCOL_ERROR, None
        else:
            if mode > 1:
                self._logger.error(f"Unknown trigger mode: {mode}, it must be one of (0, 1)")
                return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">B", mode)
        self._logger.debug(f"osc_trigger_mode payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_TRIGGER_MODE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_trigger_mode = mode

        return status, None

    def osc_trigger_source(self, source: int | str) -> tuple[int, None]:
        """
        Set trigger source.

        :param source: Trigger source: It can be one of ('N', 'A', 'B', 'P') or ('Nut'、'ChA'、'ChB'、'Protocol')
                       or a number in 0, 1, 2, 3, represent Nut, Channel A, Channel B, and Protocol, respectively.
        :type source: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        sources1 = ("N", "A", "B", "P")
        sources2 = ("NUT", "CHA", "CHB", "PROTOCOL")
        if isinstance(source, str):
            source = source.upper()
            if source in sources1:
                source = sources1.index(source)
            elif source in sources2:
                source = sources2.index(source)
            else:
                self._logger.error(
                    f"Invalid trigger source: {source}. it must be one of {sources1} or {sources2} "
                    f"if specified as a string."
                )
                return self.NON_PROTOCOL_ERROR, None
        else:
            if source > 3:
                self._logger.error("Invalid trigger source, it must be one of (0, 1, 2, 3)")
                return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">B", source)
        self._logger.debug(f"osc_trigger_source payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_ANALOG_TRIGGER_SOURCE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_analog_trigger_source = source

        return status, None

    def osc_trigger_edge(self, edge: int | str) -> tuple[int, None]:
        """
        Set trigger edge.

        :param edge: Trigger edge. ('up', 'down', 'either') or ('u', 'd', 'e') or 0, 1, 2
                     represent up, down, either, respectively.
        :type edge: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        edges1 = ("UP", "DOWN", "EITHER")
        edges2 = ("U", "D", "E")
        if isinstance(edge, str):
            edge = edge.upper()
            if edge in edges1:
                edge = edges1.index(edge)
            elif edge in edges2:
                edge = edges2.index(edge)
            else:
                self._logger.error(
                    f"Invalid trigger edge: {edge}. it must be one of {edges1} or {edges2} "
                    f"if specified as a string."
                )
                return self.NON_PROTOCOL_ERROR, None
        if edge not in (0, 1, 2):
            self._logger.error("Invalid trigger edge, it must be one of (0, 1, 2)")
            return self.NON_PROTOCOL_ERROR, None

        payload = struct.pack(">B", edge)
        self._logger.debug(f"osc_trigger_edge payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_TRIGGER_EDGE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_analog_trigger_edge = edge

        return status, None

    def osc_trigger_level(self, edge_level: int) -> tuple[int, None]:
        """
        Set trigger edge level.

        :param edge_level: Edge level.
        :type edge_level: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">h", edge_level)
        self._logger.debug(f"osc_trigger_level payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_TRIGGER_EDGE_LEVEL, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_analog_trigger_edge_level = edge_level

        return status, None

    def osc_sample_delay(self, delay: int | str) -> tuple[int, None]:
        """
        Set sample delay.

        :param delay: Sample delay. It can be a number or a string, if the input is a string,
                       you can specify the unit, ending with the value,
                       using either 'k' (representing 1024) or 'm' (representing 1024^2).
                       For example, "1k" means 1024 and "2m" means 2097152.
        :type delay: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(delay, str):
            m = re.match(r"^(\d+)([kKmM])?$", delay)

            if m:
                delay = int(m.group(1))
            else:
                self._logger.error(
                    "Input format error, it should be a number or a string "
                    "ending with the value, using either k or m."
                )
                return self.NON_PROTOCOL_ERROR, None

            if m.group(2):
                unit = m.group(2).lower()
                if unit == "k":
                    delay *= 1024
                elif unit == "m":
                    delay *= 1024 * 1024
                else:
                    self._logger.error(f"Unknown unit: {unit}")
                    return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">i", delay)
        self._logger.debug(f"osc_sample_delay payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_SAMPLE_DELAY, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_delay = delay

        return status, None

    def osc_sample_length(self, length: int | str) -> tuple[int, None]:
        """
        Set sample length.

        :param length: Sample length. It can be a number or a string, if the input is a string,
                       you can specify the unit, ending with the value,
                       using either 'k' (representing 1024) or 'm' (representing 1024^2).
                       For example, "1k" means 1024 and "2m" means 2097152.
        :type length: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(length, str):
            m = re.match(r"^(\d+)([kKmM])?$", length)

            if m:
                length = int(m.group(1))
            else:
                self._logger.error(
                    "Input format error, it should be a number or a string "
                    "ending with the value, using either k or m."
                )
                return self.NON_PROTOCOL_ERROR, None

            if m.group(2):
                unit = m.group(2).lower()
                if unit == "k":
                    length *= 1024
                elif unit == "m":
                    length *= 1024 * 1024
                else:
                    self._logger.error(f"Unknown unit: {unit}")
                    return self.NON_PROTOCOL_ERROR, None

        payload = struct.pack(">I", length)
        self._logger.debug(f"osc_sample_length payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_SAMPLE_LENGTH, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_length = length

        return status, None

    def osc_sample_clock(self, clock: int | str) -> tuple[int, None]:
        """
        Set osc sample rate

        :param clock: The sample rate in kHz can be one of (65000, 48000, 24000, 12000, 8000, 4000)
                      or a string in (65M, 48M, 24M, 12M, 8M, 4M).
        :type clock: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(clock, str):
            clock = clock.upper()
            if clock == "65M":
                clock = 65000
            elif clock == "48M":
                clock = 48000
            elif clock == "24M":
                clock = 24000
            elif clock == "12M":
                clock = 12000
            elif clock == "8M":
                clock = 8000
            elif clock == "4M":
                clock = 4000
            else:
                if re.match(r"^\d+$", clock):
                    clock = int(clock)
                else:
                    self._logger.error("UnSupport osc sample rate, it should in 65M or 48M or 24M or 12M or 8M or 4M")
                    return self.NON_PROTOCOL_ERROR, None
        if clock not in (65000, 48000, 24000, 12000, 8000, 4000):
            self._logger.error("UnSupport osc sample clock, it should in (65000, 48000, 24000, 12000, 8000, 4000)")
            return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">I", clock)
        self._logger.debug(f"osc_sample_clock_rate payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_SAMPLE_RATE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_clock = clock

        return status, None

    def osc_analog_gain(self, channel: int | str, gain: int) -> tuple[int, None]:
        """
        Set analog gain.

        :param channel: Analog channel, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :param gain: Analog gain.
        :type gain: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        channels = ("A", "B")
        if isinstance(channel, str):
            channel = channel.upper()
            if channel in channels:
                channel = channels.index(channel)
            else:
                self._logger.error("Channel error, it should in 'A' or 'B'")
                return self.NON_PROTOCOL_ERROR, None
        else:
            if channel > 1:
                self._logger.error("Channel error, it should in 0 or 1")
                return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">BB", channel, gain)
        self._logger.debug(f"osc_analog_gain payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_ANALOG_GAIN, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_analog_gain[channel] = gain

        return status, None

    def osc_sample_clock_phase(self, phase: int) -> tuple[int, None]:
        """
        Set sample phase.

        :param phase: Sample phase.
        :type phase: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">I", phase)
        self._logger.debug(f"osc_sample_clock_phase payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_CLOCK_SAMPLE_PHASE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_phase = phase

        return status, None

    def nut_voltage_disable(self) -> tuple[int, None]:
        """
        Disable the nut voltage.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_enable(False)

    def nut_voltage_enable(self) -> tuple[int, None]:
        """
        Enable the nut voltage.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_enable(True)

    def _nut_set_enable(self, enable: int | bool) -> tuple[int, None]:
        """
        Set nut enable.

        :param enable: Enable or disable. 0 for disable, 1 for enable if specified as a integer.
        :type enable: int | bool
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(enable, int):
            enable = True if enable == 1 else False
        payload = struct.pack(">?", enable)
        self._logger.debug(f"nut_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_enable = enable

        return status, None

    def nut_voltage(self, voltage: float | str | int) -> tuple[int, None]:
        """
        Set nut voltage.

        :param voltage: Nut voltage can be a number or a string. If the input is a float,
                        the voltage will be treated as a V value. If the input is a string,
                        you can specify the unit, ending with the value, using either mV or V.
        :type voltage: float | str | int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(voltage, str):
            m = re.match(r"^(\d+(?:\.\d+)?)([mM]?[vV])?$", voltage)
            if not m:
                self._logger.error(
                    "Input format error: " "it should be a number or a string and end with a unit in V or mV."
                )
                return self.NON_PROTOCOL_ERROR, None
            voltage = float(m.group(1))
            if m.group(2):
                unit = m.group(2)
            else:
                unit = "v"
            if unit.lower() == "v":
                voltage = voltage * 1000
        else:
            voltage = voltage * 1000
        voltage = int(voltage)
        payload = struct.pack(">I", voltage)
        self._logger.debug(f"nut_voltage payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_VOLTAGE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_voltage = voltage / 1000

        return status, None

    def nut_clock_disable(self) -> tuple[int, None]:
        """
        Disable the nut clock.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_clock_enable(False)

    def nut_clock_enable(self) -> tuple[int, None]:
        """
        Enable the nut clock.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_clock_enable(True)

    def _nut_set_clock_enable(self, enable: bool) -> tuple[int, None]:
        payload = struct.pack(">?", enable)
        self._logger.debug(f"nut_set_clock_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_CLOCK_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_clock_enable = enable

        return status, None

    def nut_clock_freq(self, clock: int | str) -> tuple[int, None]:
        """
        Set nut clock.

        :param clock: The clock of the nut in kHz
        :type clock: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(clock, str):
            clock = clock.upper()
            if clock == "24M":
                clock = 24000
            elif clock == "12M":
                clock = 12000
            elif clock == "8M":
                clock = 8000
            elif clock == "4M":
                clock = 4000
            else:
                self._logger.error(f"Unknown clock type: {clock}, 24M or 12M or 8M or 4M")
                return protocol.STATUS_ERROR, None

        payload = struct.pack(">I", clock)
        self._logger.debug(f"nut_set_clock payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_CLOCK, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_clock = clock

        return status, None

    def nut_timeout_ms(self, timeout: int) -> tuple[int, None]:
        """
        Set nut timeout.

        :param timeout: Nut timeout.
        :type timeout: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">I", timeout)
        self._logger.debug(f"nut_timeout_ms payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_TIMEOUT, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_timeout = timeout

        return status, None

    def spi_enable(self) -> tuple[int, None]:
        """
        Enable the SPI

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._spi_enable(True)

    def spi_disable(self) -> tuple[int, None]:
        """
        Disable the SPI

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._spi_enable(False)

    def _spi_enable(self, enable: bool):
        """
        Enable the SPI.

        :param enable: True for enable, False for disable.
        :type enable: bool
        :return: The device response status.
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">?", enable)
        self._logger.debug(f"cracker_spi_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_SPI_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self.get_current_config().cracker_uart_enable = enable
        return status, res

    def spi_reset(self) -> tuple[int, None]:
        """
        Reset the SPI hardware.

        :return: The device response status.
        :rtype: tuple[int, None]
        """
        payload = None
        self._logger.debug(f"cracker_spi_reset payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_SPI_RESET)

    def spi_config(
        self,
        speed: int = 10_000,
        cpol: serial.SpiCpol = serial.SpiCpol.SPI_CPOL_LOW,
        cpha: serial.SpiCpha = serial.SpiCpha.SPI_CPHA_LOW,
    ) -> tuple[int, None]:
        """
        Config the SPI.

        :param speed: SPI speed.
        :type speed: int
        :param cpol: Clock polarity.
        :type cpol: serial.SpiCpol
        :param cpha: Clock phase.
        :type cpha: serial.SpiCpha
        :return: The device response status.
        :rtype: tuple[int, None]
        """

        # System clock is 100e6
        # Clock divider is 2
        # psc max is 65535
        psc = 100e6 / 2 / speed
        if psc > 65535 or psc < 2:
            return protocol.STATUS_COMMAND_UNSUPPORTED, None

        if not psc.is_integer():
            _psc = psc
            psc = round(psc)
            if psc > 65535 or psc < 2:
                return protocol.STATUS_COMMAND_UNSUPPORTED, None
            self._logger.warning(
                f"The speed: [{speed}] cannot calculate an integer Prescaler, " f"so the integer value is set to {psc}."
            )

        payload = struct.pack(">HBB", int(psc), cpol.value, cpha.value)
        self._logger.debug(f"cracker_spi_config payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_SPI_CONFIG, payload=payload)
        if status == protocol.STATUS_OK:
            self.get_current_config().cracker_i2c_config = {
                "speed": speed,
                "cpol": cpol,
                "cpha": cpha,
            }
        return status, res

    def _spi_transceive(
        self, tx_data: bytes | str | None, is_delay: bool, delay: int, rx_count: int, is_trigger: bool
    ) -> tuple[int, bytes | None]:
        """
        Basic interface for sending and receiving data through the SPI protocol.

        :param tx_data: The data to send.
        :param is_delay: Whether the transmit delay is enabled.
        :type is_delay: bool
        :param delay: The transmit delay in milliseconds, with a minimum effective duration of 10 nanoseconds.
        :type delay: int
        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the SPI device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)
        payload = struct.pack(">?IH?", is_delay, delay, rx_count, is_trigger)
        if tx_data is not None:
            payload += tx_data
        self._logger.debug(f"_spi_transceive payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_SPI_TRANSCEIVE, payload=payload)
        if status != protocol.STATUS_OK:
            self._logger.error(f"Receive status code error [{status}]")
            return status, None
        else:
            return status, res

    def spi_transmit(self, tx_data: bytes | str, is_trigger: bool = False) -> tuple[int, None]:
        """
        Send data through the SPI protocol.

        :param tx_data: The data to send.
        :type tx_data: str | bytes
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the SPI device.
        :rtype: tuple[int, None]
        """
        status, _ = self._spi_transceive(
            tx_data, is_delay=False, delay=1_000_000_000, rx_count=0, is_trigger=is_trigger
        )
        return status, None

    def spi_receive(self, rx_count: int, is_trigger: bool = False) -> tuple[int, bytes | None]:
        """
        Receive data through the SPI protocol.

        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the SPI device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        return self._spi_transceive(None, is_delay=False, delay=1_000_000_000, rx_count=rx_count, is_trigger=is_trigger)

    def spi_transmit_delay_receive(
        self, tx_data: bytes | str, delay: int, rx_count: int, is_trigger: bool = False
    ) -> tuple[int, bytes | None]:
        """
        Send and receive data with delay through the SPI protocol.

        :param tx_data: The data to send.
        :type tx_data: str | bytes
        :param delay: The transmit delay in milliseconds, with a minimum effective duration of 10 nanoseconds.
        :type delay: int
        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the SPI device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        return self._spi_transceive(tx_data, is_delay=True, delay=delay, rx_count=rx_count, is_trigger=is_trigger)

    def spi_transceive(self, tx_data: bytes | str, rx_count: int, is_trigger: bool = False) -> tuple[int, bytes | None]:
        """
        Send and receive data without delay through the SPI protocol.

        :param tx_data: The data to send.
        :type tx_data: str | bytes
        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the SPI device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        return self._spi_transceive(tx_data, is_delay=False, delay=0, rx_count=rx_count, is_trigger=is_trigger)

    def i2c_enable(self) -> tuple[int, None]:
        """
        Enable the I2C

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._i2c_enable(True)

    def i2c_disable(self) -> tuple[int, None]:
        """
        Disable the I2C

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._i2c_enable(False)

    def _i2c_enable(self, enable: bool):
        """
        Enable the I2C.

        :param enable: True for enable, False for disable.
        :type enable: bool
        :return: The device response status.
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">?", enable)
        self._logger.debug(f"cracker_i2c_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_I2C_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self.get_current_config().cracker_uart_enable = enable
        return status, res

    def i2c_reset(self) -> tuple[int, None]:
        """
        Reset the I2C.

        :return: The device response status.
        :rtype: tuple[int, None]
        """
        payload = None
        self._logger.debug(f"cracker_i2c_reset payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_I2C_RESET)

    def i2c_config(
        self,
        dev_addr: int = 0x00,
        speed: serial.I2cSpeed = serial.I2cSpeed.STANDARD_100K,
    ) -> tuple[int, None]:
        """
        Config the SPI.

        :param dev_addr: The address of the device.
        :type dev_addr: int
        :param speed: The speed of the device.
        :type speed: serial.I2cSpeed
        :return: The device response status.
        :rtype: tuple[int, None]
        """

        payload = struct.pack(">BB", dev_addr, speed.value)
        self._logger.debug(f"cracker_i2c_config payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_I2C_CONFIG, payload=payload)
        if status == protocol.STATUS_OK:
            self.get_current_config().cracker_i2c_config = {
                "dev_addr": dev_addr,
                "speed": speed,
            }
        return status, res

    def _i2c_transceive(
        self,
        tx_data: bytes | str | None,
        combined_transfer_count_1: int,
        combined_transfer_count_2: int,
        transfer_rw: tuple[int, int, int, int, int, int, int, int],
        transfer_lens: tuple[int, int, int, int, int, int, int, int],
        is_delay: bool,
        delay: int,
        is_trigger: bool,
    ) -> tuple[int, bytes | None]:
        """
        Basic API for sending and receiving data through the I2C protocol.

        :param tx_data: The data to be sent.
        :type tx_data: bytes | str | None
        :param combined_transfer_count_1: The first combined transmit transfer count.
        :type combined_transfer_count_1: int
        :param combined_transfer_count_2: The second combined transmit transfer count.
        :type combined_transfer_count_2: int
        :param transfer_rw: The read/write configuration tuple of the four transfers in the two sets
                            of Combined Transfer, with a tuple length of 8, where 0 represents write
                            and 1 represents read.
        :type transfer_rw: tuple[int, int, int, int, int, int, int, int, int]
        :param transfer_lens: The transfer length tuple of the four transfers in the two combined transmit sets.
        :type transfer_lens: tuple[int, int, int, int, int, int, int, int]
        :param is_delay: Whether the transmit delay is enabled.
        :type is_delay: bool
        :param delay: Transmit delay duration, in nanoseconds, with a minimum effective duration of 10 nanoseconds.
        :type delay: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the I2C device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)

        if combined_transfer_count_1 > 4:
            raise ValueError("Illegal combined combined_transfer_count_1")
        if combined_transfer_count_2 > 4:
            raise ValueError("Illegal combined combined_transfer_count_2")

        if len(transfer_rw) != 8:
            raise ValueError("transfer_rw length must be 8")
        if len(transfer_lens) != 8:
            raise ValueError("transfer_lens length must be 8")

        transfer_rw_num = sum(bit << (7 - i) for i, bit in enumerate(transfer_rw))

        payload = struct.pack(
            ">?I3B8H?",
            is_delay,
            delay,
            combined_transfer_count_1,
            combined_transfer_count_2,
            transfer_rw_num,
            *transfer_lens,
            is_trigger,
        )

        if tx_data is not None:
            payload += tx_data
        status, res = self.send_with_command(protocol.Command.CRACKER_I2C_TRANSCEIVE, payload=payload)
        if status != protocol.STATUS_OK:
            self._logger.error(f"Receive status code error [{status}]")
            return status, None
        else:
            return status, res

    def i2c_transmit(self, tx_data: bytes | str, is_trigger: bool = False) -> tuple[int, None]:
        """
        Send data through the I2C protocol.

        :param tx_data: The data to be sent.
        :type tx_data: str | bytes
        :param is_trigger: Whether the transmit trigger is enabled.
        """
        transfer_rw = (0, 0, 0, 0, 0, 0, 0, 0)
        transfer_lens = (len(tx_data), 0, 0, 0, 0, 0, 0, 0)
        status, _ = self._i2c_transceive(
            tx_data,
            combined_transfer_count_1=1,
            combined_transfer_count_2=0,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=False,
            delay=0,
            is_trigger=is_trigger,
        )
        return status, None

    def i2c_receive(self, rx_count, is_trigger: bool = False) -> tuple[int, bytes | None]:
        """
        Receive data through the I2C protocol.

        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :return: The device response status and the data received from the I2C device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        transfer_rw = (1, 1, 1, 1, 1, 1, 1, 1)
        transfer_lens = (rx_count, 0, 0, 0, 0, 0, 0, 0)
        return self._i2c_transceive(
            tx_data=None,
            combined_transfer_count_1=1,
            combined_transfer_count_2=0,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=False,
            delay=0,
            is_trigger=is_trigger,
        )

    def i2c_transmit_delay_receive(
        self, tx_data: bytes | str, delay: int, rx_count: int, is_trigger: bool = False
    ) -> tuple[int, bytes | None]:
        """
        Send and receive data with delay through the I2C protocol.

        :param tx_data: The data to be sent.
        :type tx_data: str | bytes
        :param delay: Transmit delay duration, in nanoseconds, with a minimum effective duration of 10 nanoseconds.
        :type delay: int
        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the I2C device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        transfer_rw = (0, 0, 0, 0, 1, 1, 1, 1)
        transfer_lens = (len(tx_data), 0, 0, 0, rx_count, 0, 0, 0)
        return self._i2c_transceive(
            tx_data,
            combined_transfer_count_1=1,
            combined_transfer_count_2=1,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=True,
            delay=delay,
            is_trigger=is_trigger,
        )

    def i2c_transceive(self, tx_data, rx_count, is_trigger: bool = False) -> tuple[int, bytes | None]:
        """
        Send and receive data without delay through the I2C protocol.

        :param tx_data: The data to be sent.
        :type tx_data: str | bytes
        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the I2C device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        transfer_rw = (0, 0, 0, 0, 1, 1, 1, 1)
        transfer_lens = (len(tx_data), 0, 0, 0, rx_count, 0, 0, 0)
        return self._i2c_transceive(
            tx_data,
            combined_transfer_count_1=1,
            combined_transfer_count_2=1,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=False,
            delay=0,
            is_trigger=is_trigger,
        )

    def uart_enable(self) -> tuple[int, None]:
        """
        Enable the UART.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._uart_enable(True)

    def uart_disable(self) -> tuple[int, None]:
        """
        Disable the UART.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._uart_enable(False)

    def _uart_enable(self, enable: bool) -> tuple[int, None]:
        """
        Enable the uart.

        :param enable: True for enable, False for disable.
        :type enable: bool
        :return: The device response status.
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">?", enable)
        self._logger.debug(f"cracker_uart_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_UART_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self.get_current_config().cracker_uart_enable = enable
        return status, res

    def uart_reset(self) -> tuple[int, None]:
        """
        Reset the UART hardware.

        :return: The device response status.
        :rtype: tuple[int, None]
        """
        payload = None
        self._logger.debug(f"cracker_uart_reset payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_UART_RESET)

    def uart_config(
        self,
        baudrate: serial.Baudrate = serial.Baudrate.BAUDRATE_115200,
        bytesize: serial.Bytesize = serial.Bytesize.EIGHTBITS,
        parity: serial.Parity = serial.Parity.PARITY_NONE,
        stopbits: serial.Stopbits = serial.Stopbits.STOPBITS_ONE,
    ) -> tuple[int, None]:
        """
        Config uart.

        :param baudrate: The baudrate of the uart.
        :type baudrate: serial.Baudrate
        :param bytesize: The bytesize of the uart.
        :type bytesize: serial.Bytesize
        :param parity: The parity of the uart.
        :type parity: serial.Parity
        :param stopbits: The stopbits of the uart.
        :type stopbits: serial.Stopbits
        :return: The device response status.
        :rtype: tuple[int, None]
        """

        payload = struct.pack(">BBBB", stopbits.value, parity.value, bytesize.value, baudrate.value)
        self._logger.debug(f"cracker_uart_config payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_UART_CONFIG, payload=payload)
        if status == protocol.STATUS_OK:
            self.get_current_config().cracker_uart_config = {
                "baudrate": baudrate,
                "bytesize": bytesize,
                "parity": parity,
                "stopbits": stopbits,
            }
        return status, res

    def uart_transmit_receive(
        self, tx_data: str | bytes = None, rx_count: int = 0, is_trigger: bool = False, timeout: int = 10000
    ) -> tuple[int, bytes | None]:
        """
        Transmit and receive data through the UART protocol.

        :param tx_data: The data to be sent.
        :type tx_data: str | bytes
        :param rx_count: The number of received data bytes.
        :type rx_count: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :param timeout: Timeout in milliseconds.
        :type timeout: int
        :return: The device response status and the data received from the device.
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)

        payload = struct.pack(">H?I", rx_count, is_trigger, timeout)
        if tx_data is not None:
            payload += tx_data
        self._logger.debug(f"cracker_uart_transmit_receive payload: {payload.hex()}")
        return self.send_with_command(protocol.Command.CRACKER_UART_TRANSCEIVE, payload=payload)

    def uart_receive_fifo_remained(self) -> tuple[int, int]:
        """
        Get the number of remaining unread bytes in the UART receive FIFO.

        :return: The device response status and the number of remaining unread bytes.
        :rtype: tuple[int, int]
        """
        payload = None
        self._logger.debug(f"cracker_uart_receive_fifo_remained payload: {payload}")
        status, res = self.send_with_command(protocol.Command.CRACKER_UART_RECEIVE_FIFO_REMAINED)
        return status, struct.unpack(">H", res)[0]

    def uart_receive_fifo_dump(self) -> tuple[int, bytes | None]:
        """
        Read all the remaining data from the UART receive FIFO.

        :return: The device response status and all the remaining unread bytes.
        :rtype: tuple[int, bytes | None]
        """
        payload = None
        self._logger.debug(f"cracker_uart_receive_fifo_dump payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_UART_CRACKER_UART_RECEIVE_FIFO_DUMP)

    def uart_receive_fifo_clear(self) -> tuple[int, bytes | None]:
        """
        Clear all the remaining data in the UART receive FIFO.

        :return: The device response status and all the remaining unread bytes.
        :rtype: tuple[int, bytes | None]
        """
        payload = None
        self._logger.debug(f"cracker_uart_receive_fifo_dump payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_UART_CRACKER_UART_RECEIVE_CLEAR)



# Copyright 2024 CrackNuts. All rights reserved.

import pathlib
import threading
import time
import typing

import numpy as np
import traitlets
import cracknuts.logger as logger
from cracknuts.acquisition.acquisition import Acquisition
from cracknuts.jupyter.panel import MsgHandlerPanelWidget
from cracknuts.scope.scope_acquisition import ScopeAcquisition


class ScopePanelWidget(MsgHandlerPanelWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "ScopePanelWidget.js"
    _css = ""

    series_data = traitlets.Dict({}).tag(sync=True)

    custom_y_range: dict[str, tuple[int, int]] = traitlets.Dict({"1": (0, 0), "2": (0, 0)}).tag(sync=True)
    y_range: dict[int, tuple[int, int]] = traitlets.Dict({1: (None, None), 2: (None, None)}).tag(sync=True)
    combine_y_range = traitlets.Bool(False).tag(sync=True)

    scope_status = traitlets.Int(0).tag(sync=True)
    monitor_status = traitlets.Bool(False).tag(sync=True)
    lock_scope_operation = traitlets.Bool(False).tag(sync=True)

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(*args, **kwargs)
        self._logger = logger
        if not hasattr(self, "acquisition"):
            self._acquisition: Acquisition | None = None
            if "acquisition" in kwargs and isinstance(kwargs["acquisition"], Acquisition):
                self._acquisition = kwargs["acquisition"]
            if self._acquisition is None:
                raise ValueError("acquisition is required")

        # 0 means scope_acquisition is effective, 1 means acquisition is effective,
        # and by default, scope_acquisition is effective.
        self._effect_acq = 0
        self._scope_acquisition = ScopeAcquisition(self._acquisition.cracker)
        self._acquisition.on_status_changed(self._change_acquisition_source)
        self._scope_acquisition.on_status_changed(self._change_scope_acquisition_status)
        self._monitor_period = 0.1

    def _change_acquisition_source(self, status: int) -> None:
        # Listen the acquisition thread status change and update scope monitor status.
        if status == 0:
            self.lock_scope_operation = False
        else:
            self._effect_acq = 1
            self.lock_scope_operation = True
            self.scope_status = 0
            if not self.monitor_status:
                self.start_monitor()

    def _change_scope_acquisition_status(self, status: int) -> None:
        # Only listen the stop change from scope acquisition thread, and sync status to ui.
        if status == 0 and self._effect_acq == 0:
            self.scope_status = status

    def _update_status(self, status: int) -> None:
        self.scope_status = status

    def update(self, series_data: dict[int, np.ndarray]) -> None:
        (
            mn1,
            mx1,
        ) = None, None
        (
            mn2,
            mx2,
        ) = None, None

        if 1 in series_data.keys():
            c1 = series_data[1]
            mn1, mx1 = np.min(c1), np.max(c1)
        if 2 in series_data.keys():
            c2 = series_data[2]
            mn2, mx2 = np.min(c2), np.max(c2)

        self.y_range = {1: (mn1, mx1), 2: (mn2, mx2)}

        self.series_data = {k: v.tolist() for k, v in series_data.items()}

    @traitlets.observe("scope_status")
    def scope_status_changed(self, change) -> None:
        self.run(change.get("new"))

    @traitlets.observe("monitor_status")
    def monitor_status_changed(self, change) -> None:
        if change.get("new"):
            self.start_monitor()

    def run(self, status: int) -> None:
        if not self.lock_scope_operation:
            self._scope_acquisition.run(status)
            self._effect_acq = 0
            if not self.monitor_status:
                self.start_monitor()

    def _monitor(self) -> None:
        while self.monitor_status:
            if self._effect_acq == 0:
                wave = self._scope_acquisition.get_last_wave()
                if not self._scope_acquisition.is_running():
                    self.monitor_status = False
            else:
                if self._scope_acquisition.is_running():
                    self._scope_acquisition.stop()
                wave = self._acquisition.get_last_wave()
                if not self._acquisition.is_running():
                    self.monitor_status = False
            self.update(wave)
            time.sleep(self._monitor_period)

    def start_monitor(self) -> None:
        self.monitor_status = True
        threading.Thread(target=self._monitor).start()

    def stop_monitor(self) -> None:
        self.monitor_status = False



# Copyright 2024 CrackNuts. All rights reserved.

import pathlib
import typing
from typing import Any

from traitlets import traitlets

from cracknuts import logger
from cracknuts.cracker.cracker_s1 import CrackerS1
from cracknuts.jupyter.panel import MsgHandlerPanelWidget
from cracknuts.jupyter.ui_sync import ConfigProxy, observe_interceptor
import cracknuts.cracker.serial as serial


class CrackerS1PanelWidget(MsgHandlerPanelWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "CrackerS1PanelWidget.js"
    _css = ""

    uri = traitlets.Unicode("cnp://192.168.0.11:8080").tag(sync=True)
    connect_status = traitlets.Bool(False).tag(sync=True)
    cracker_id = traitlets.Unicode("Unknown").tag(sync=True)
    cracker_name = traitlets.Unicode("Unknown").tag(sync=True)
    cracker_version = traitlets.Unicode("Unknown").tag(sync=True)

    # nut
    nut_enable = traitlets.Bool(False).tag(sync=True)
    nut_voltage = traitlets.Float(3.3).tag(sync=True)
    nut_clock_enable = traitlets.Bool(False).tag(sync=True)
    nut_clock = traitlets.Int(65000).tag(sync=True)

    nut_uart_enable = traitlets.Bool(False).tag(sync=True)
    nut_uart_baudrate = traitlets.Int(115200).tag(sync=True)
    nut_uart_bytesize = traitlets.Int(8).tag(sync=True)
    nut_uart_parity = traitlets.Int(0).tag(sync=True)
    nut_uart_stopbits = traitlets.Int(0).tag(sync=True)

    nut_spi_enable = traitlets.Bool(False).tag(sync=True)
    nut_spi_speed = traitlets.Int(10_000).tag(sync=True)
    nut_spi_cpol = traitlets.Int(0).tag(sync=True)
    nut_spi_cpha = traitlets.Int(0).tag(sync=True)

    nut_i2c_enable = traitlets.Bool(False).tag(sync=True)
    nut_i2c_dev_addr = traitlets.Unicode("0x00").tag(sync=True)
    nut_i2c_speed = traitlets.Int(0).tag(sync=True)

    # osc
    osc_analog_channel_a_enable = traitlets.Bool(False).tag(sync=True)
    osc_analog_channel_b_enable = traitlets.Bool(True).tag(sync=True)
    sync_sample = traitlets.Bool(False).tag(sync=True)
    sync_args_times = traitlets.Int(1).tag(sync=True)

    osc_sample_rate = traitlets.Int(65000).tag(sync=True)
    osc_sample_phase = traitlets.Int(0).tag(sync=True)
    osc_sample_len = traitlets.Int(1024).tag(sync=True)
    osc_sample_delay = traitlets.Int(1024).tag(sync=True)

    osc_trigger_source = traitlets.Int(0).tag(sync=True)
    osc_trigger_mode = traitlets.Int(0).tag(sync=True)
    osc_trigger_edge = traitlets.Int(0).tag(sync=True)
    osc_trigger_edge_level = traitlets.Int(0).tag(sync=True)

    osc_analog_channel_a_gain = traitlets.Int(1).tag(sync=True)
    osc_analog_channel_b_gain = traitlets.Int(1).tag(sync=True)

    def __init__(self, *args: Any, **kwargs: Any):
        # todo optimize init param with specifically args and kwargs.
        super().__init__(*args, **kwargs)
        self._logger = logger.get_logger(self)
        self._observe: bool = True
        self.cracker: CrackerS1 | None = None
        if "cracker" in kwargs:
            self.cracker: CrackerS1 = kwargs["cracker"]
        if self.cracker is None:
            raise ValueError("cracker is required")
        self.reg_msg_handler("connectButton", "onClick", self.msg_connection_button_on_click)
        self.connect_status = self.cracker.get_connection_status()
        if self.connect_status:
            _, self.cracker_id = self.cracker.get_id()
            _, self.cracker_name = self.cracker.get_hardware_model()
            _, self.cracker_version = self.cracker.get_firmware_version()

    def sync_config(self) -> None:
        """
        Sync cracker current to panel(Jupyter widget UI)
        """
        # connection
        self._observe = False
        if self.cracker.get_uri() is not None:
            self.uri = self.cracker.get_uri()

        current_config = self.cracker.get_current_config()

        # nut
        if current_config.nut_enable is not None:
            self.nut_enable = current_config.nut_enable
        if current_config.nut_voltage is not None:
            self.nut_voltage = current_config.nut_voltage
        if current_config.nut_clock is not None:
            self.nut_clock = current_config.nut_clock
        if current_config.nut_clock_enable is not None:
            self.nut_clock_enable = current_config.nut_clock_enable

        # osc
        self.osc_analog_channel_a_enable = current_config.osc_analog_channel_enable.get(0, False)
        self.osc_analog_channel_b_enable = current_config.osc_analog_channel_enable.get(1, True)
        self.osc_analog_channel_a_gain = current_config.osc_analog_gain.get(0, 1)
        self.osc_analog_channel_b_gain = current_config.osc_analog_gain.get(1, 1)
        if current_config.osc_sample_length is not None:
            self.osc_sample_len = current_config.osc_sample_length
        if current_config.osc_sample_delay is not None:
            self.osc_sample_delay = current_config.osc_sample_delay
        if current_config.osc_sample_clock is not None:
            self.osc_sample_rate = current_config.osc_sample_clock
        if current_config.osc_sample_phase is not None:
            self.osc_sample_phase = current_config.osc_sample_phase
        if current_config.osc_analog_trigger_source is not None:
            self.osc_trigger_source = current_config.osc_analog_trigger_source
        if current_config.osc_trigger_mode is not None:
            self.osc_trigger_mode = current_config.osc_trigger_mode
        if current_config.osc_analog_trigger_edge is not None:
            self.osc_trigger_edge = current_config.osc_analog_trigger_edge
        if current_config.osc_analog_trigger_edge_level is not None:
            self.osc_trigger_edge_level = current_config.osc_analog_trigger_edge_level

        self._observe = True

    def bind(self) -> None:
        """
        Bind the cracker and crackerPanel objects so that when the configuration of cracker is set,
        the updated values are automatically synchronized with panel through a ProxyConfig object.
        """
        proxy_config = ConfigProxy(self.cracker.get_current_config(), self)
        self.cracker._config = proxy_config

    def msg_connection_button_on_click(self, args: dict[str, typing.Any]):
        if args.get("action") == "connect":
            self.cracker.connect()
            if self.cracker.get_connection_status():
                self.connect_status = True
                _, self.cracker_id = self.cracker.get_id()
                _, self.cracker_name = self.cracker.get_hardware_model()
                _, self.cracker_version = self.cracker.get_firmware_version()
            else:
                self.connect_status = False
        else:
            self.cracker.disconnect()
            self.connect_status = False
        self.send({"connectFinished": self.connect_status})

    @traitlets.observe("uri")
    @observe_interceptor
    def uri_on_changed(self, change):
        self.cracker.set_uri(change.get("new"))

    @traitlets.observe("nut_enable")
    @observe_interceptor
    def nut_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.nut_voltage_enable() if enabled else self.cracker.nut_voltage_disable()
        if enabled:
            self.cracker.nut_voltage(self.cracker.get_current_config().nut_voltage)

    @traitlets.observe("nut_voltage")
    @observe_interceptor
    def nut_voltage_changed(self, change):
        self.cracker.nut_voltage(change.get("new"))

    @traitlets.observe("nut_clock_enable")
    @observe_interceptor
    def nut_clock_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.nut_clock_enable() if enabled else self.cracker.nut_clock_disable()
        if enabled:
            self.cracker.nut_clock_freq(self.cracker.get_current_config().nut_clock)

    @traitlets.observe("nut_clock")
    @observe_interceptor
    def nut_clock_changed(self, change):
        self.cracker.nut_clock_freq(int(change.get("new")))

    @traitlets.observe("osc_sample_phase")
    @observe_interceptor
    def osc_sample_phase_changed(self, change):
        self.cracker.osc_sample_clock_phase(int(change.get("new")))

    @traitlets.observe("osc_sample_len")
    @observe_interceptor
    def osc_sample_len_changed(self, change):
        self.cracker.osc_sample_length(int(change.get("new")))

    @traitlets.observe("osc_sample_delay")
    @observe_interceptor
    def osc_sample_delay_changed(self, change):
        self.cracker.osc_sample_delay(int(change.get("new")))

    @traitlets.observe("osc_sample_rate")
    @observe_interceptor
    def osc_sample_rate_changed(self, change):
        self.cracker.osc_sample_clock(int(change.get("new")))

    @traitlets.observe("osc_analog_channel_a_enable")
    @observe_interceptor
    def osc_analog_channel_a_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.osc_analog_enable(0) if enabled else self.cracker.osc_analog_disable(0)
        if enabled:
            self.cracker.osc_analog_gain(0, self.cracker.get_current_config().osc_analog_gain[0])

    @traitlets.observe("osc_analog_channel_b_enable")
    @observe_interceptor
    def osc_analog_channel_b_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.osc_analog_enable(1) if enabled else self.cracker.osc_analog_disable(1)
        if enabled:
            self.cracker.osc_analog_gain(1, self.cracker.get_current_config().osc_analog_gain[1])

    @traitlets.observe("osc_trigger_source")
    @observe_interceptor
    def osc_trigger_source_changed(self, change):
        self.cracker.osc_trigger_source(change.get("new"))

    @traitlets.observe("osc_trigger_mode")
    @observe_interceptor
    def osc_trigger_mode_changed(self, change):
        self.cracker.osc_trigger_mode(change.get("new"))

    @traitlets.observe("osc_trigger_edge")
    @observe_interceptor
    def osc_trigger_edge_changed(self, change):
        self.cracker.osc_trigger_edge(change.get("new"))

    @traitlets.observe("osc_trigger_edge_level")
    @observe_interceptor
    def osc_trigger_edge_level_changed(self, change):
        self.cracker.osc_trigger_level(change.get("new"))

    @traitlets.observe("osc_analog_channel_a_gain", "osc_analog_channel_b_gain")
    @observe_interceptor
    def osc_analog_channel_gain_changed(self, change):
        name = change.get("name")
        channel = None
        if name == "osc_analog_channel_a_gain":
            channel = 0
        elif name == "osc_analog_channel_b_gain":
            channel = 1
        if channel is not None:
            self.cracker.osc_analog_gain(channel, change.get("new"))

    @traitlets.observe("nut_uart_enable")
    @observe_interceptor
    def nut_uart_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.uart_enable() if enabled else self.cracker.uart_disable()
        print(f"is enabled {enabled}")
        if enabled:
            self.cracker.uart_config(
                serial.Baudrate(self.nut_uart_baudrate),
                serial.Bytesize(self.nut_uart_bytesize),
                serial.Parity(self.nut_uart_parity),
                serial.Stopbits(self.nut_uart_stopbits),
            )

    @traitlets.observe("nut_uart_baudrate")
    @observe_interceptor
    def nut_uart_baudrate_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(change.get("new")),
            serial.Bytesize(self.nut_uart_bytesize),
            serial.Parity(self.nut_uart_parity),
            serial.Stopbits(self.nut_uart_stopbits),
        )

    @traitlets.observe("nut_uart_bytesize")
    @observe_interceptor
    def nut_uart_bytesize_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(self.nut_uart_baudrate),
            serial.Bytesize(change.get("new")),
            serial.Parity(self.nut_uart_parity),
            serial.Stopbits(self.nut_uart_stopbits),
        )

    @traitlets.observe("nut_uart_parity")
    @observe_interceptor
    def nut_uart_parity_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(self.nut_uart_baudrate),
            serial.Bytesize(self.nut_uart_bytesize),
            serial.Parity(change.get("new")),
            serial.Stopbits(self.nut_uart_stopbits),
        )

    @traitlets.observe("nut_uart_stopbits")
    @observe_interceptor
    def nut_uart_stopbits_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(self.nut_uart_baudrate),
            serial.Bytesize(self.nut_uart_bytesize),
            serial.Parity(self.nut_uart_parity),
            serial.Stopbits(change.get("new")),
        )

    @traitlets.observe("nut_spi_enable")
    @observe_interceptor
    def nut_spi_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.spi_enable() if enabled else self.cracker.spi_disable()
        if enabled:
            self.cracker.spi_config(
                self.nut_spi_speed, serial.SpiCpol(self.nut_spi_cpol), serial.SpiCpha(self.nut_spi_cpha)
            )

    @traitlets.observe("nut_spi_speed")
    @observe_interceptor
    def nut_spi_speed_changed(self, change):
        self.cracker.spi_config(change.get("new"), serial.SpiCpol(self.nut_spi_cpol), serial.SpiCpha(self.nut_spi_cpha))

    @traitlets.observe("nut_spi_cpol")
    @observe_interceptor
    def nut_spi_cpol_changed(self, change):
        self.cracker.spi_config(
            self.nut_spi_speed, serial.SpiCpol(change.get("new")), serial.SpiCpha(self.nut_spi_cpha)
        )

    @traitlets.observe("nut_spi_cpha")
    @observe_interceptor
    def nut_spi_cpha_changed(self, change):
        self.cracker.spi_config(
            self.nut_spi_speed, serial.SpiCpol(self.nut_spi_cpol), serial.SpiCpha(change.get("new"))
        )

    @traitlets.observe("nut_i2c_enable")
    @observe_interceptor
    def nut_i2c_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.i2c_enable() if enabled else self.cracker.i2c_disable()
        if enabled:
            self.cracker.i2c_config(int(self.nut_i2c_dev_addr, 16), serial.I2cSpeed(self.nut_i2c_speed))

    @traitlets.observe("nut_i2c_dev_addr")
    @observe_interceptor
    def nut_i2c_dev_addr_changed(self, change):
        self.cracker.i2c_config(int(change.get("new"), 16), serial.I2cSpeed(self.nut_i2c_speed))

    @traitlets.observe("nut_i2c_speed")
    @observe_interceptor
    def nut_i2c_speed_changed(self, change):
        self.cracker.i2c_config(int(self.nut_i2c_dev_addr, 16), serial.I2cSpeed(change.get("new")))



# Copyright 2024 CrackNuts. All rights reserved.

import pathlib
import typing
from typing import Any

from cracknuts.acquisition.acquisition import Acquisition
from traitlets import traitlets

from cracknuts.jupyter.panel import MsgHandlerPanelWidget


class AcquisitionPanelWidget(MsgHandlerPanelWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "AcquisitionPanelWidget.js"
    _css = ""

    acq_status = traitlets.Int(0).tag(sync=True)
    acq_run_progress = traitlets.Dict({"finished": 0, "total": 1}).tag(sync=True)

    trace_count = traitlets.Int(1000).tag(sync=True)
    sample_offset = traitlets.Int(0).tag(sync=True)
    sample_length = traitlets.Int(1024).tag(sync=True)
    trigger_judge_wait_time = traitlets.Float(0.05).tag(sync=True)
    trigger_judge_timeout = traitlets.Float(0.005).tag(sync=True)
    do_error_max_count = traitlets.Int(1).tag(sync=True)
    file_format = traitlets.Unicode("scarr").tag(sync=True)
    file_path = traitlets.Unicode("").tag(sync=True)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        if not hasattr(self, "acquisition"):
            self.acquisition: Acquisition | None = None
            if "acquisition" in kwargs and isinstance(kwargs["acquisition"], Acquisition):
                self.acquisition: Acquisition = kwargs["acquisition"]
            if self.acquisition is None:
                raise ValueError("acquisition is required")

        self.reg_msg_handler("acqStatusButton", "onChange", self.msg_acq_status_changed)
        self.acquisition.on_status_changed(self.update_acq_status)
        self.acquisition.on_run_progress_changed(self.update_acq_run_progress)
        self.acq_status = self.acquisition.get_status()

    def sync_config(self) -> None:
        if self.acquisition.trace_count != -1:
            # When acquisition.trace_count = -1, it indicates that the ACQUISITION is in test mode,
            # and in this mode, synchronization of the acquisition trace_count to the GUI is not required.
            self.trace_count = self.acquisition.trace_count
        self.sample_offset = self.acquisition.sample_offset
        self.sample_length = self.acquisition.sample_length
        self.trigger_judge_wait_time = self.acquisition.trigger_judge_wait_time
        self.trigger_judge_timeout = self.acquisition.trigger_judge_timeout
        self.do_error_max_count = self.acquisition.do_error_max_count
        self.file_format = self.acquisition.file_format
        self.file_path = (
            ""
            if self.acquisition.file_path is None or self.acquisition.file_path == "auto"
            else self.acquisition.file_path
        )

    def bind(self) -> None:
        ...
        # todo complete ui -> python sync

    def update_acq_status(self, status) -> None:
        self.acq_status = status

    def update_acq_run_progress(self, progress: dict[str, int]):
        self.acq_run_progress = progress

    def msg_acq_status_changed(self, changed: dict[str, typing.Any]):
        status = changed.get("status")
        if status == "pause":
            self.acquisition.pause()
        elif status == "test":
            self.acquisition.test(
                sample_offset=self.sample_offset,
                trigger_judge_wait_time=self.trigger_judge_wait_time,
                trigger_judge_timeout=self.trigger_judge_timeout,
                do_error_max_count=self.do_error_max_count,
            )
        elif status == "run":
            self.acquisition.run(
                count=self.trace_count,
                sample_offset=self.sample_offset,
                sample_length=self.sample_length,
                trigger_judge_wait_time=self.trigger_judge_wait_time,
                trigger_judge_timeout=self.trigger_judge_timeout,
                do_error_max_count=self.do_error_max_count,
                file_format=self.file_format,
                file_path="auto" if self.file_path == "" or self.file_path is None else self.file_path,
            )
        else:
            self.acquisition.stop()



# Copyright 2024 CrackNuts. All rights reserved.

import abc
import json
import os.path
import time
import typing

import numpy as np
import zarr

from cracknuts import logger


class TraceDatasetData:
    def __init__(
        self,
        get_trace_data: typing.Callable[[typing.Any, typing.Any], tuple | np.ndarray],
        level: int = 0,
        index: tuple | None = None,
    ):
        self._level: int = level
        self._index: tuple | None = index
        self._get_trace_data = get_trace_data

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            index = (index,)

        level = len(index) + self._level
        index = index if self._index is None else (*self._index, *index)
        if level < 2:
            return TraceDatasetData(self._get_trace_data, level, index)
        else:
            return self._get_trace_data(*index)


class TraceDataset(abc.ABC):
    _channel_names: list[str] | None
    _channel_count: int | None
    _trace_count: int | None
    _sample_count: int | None
    _data_length: int | None
    _create_time: int | None
    _version: str | None

    @abc.abstractmethod
    def get_origin_data(self): ...

    @classmethod
    @abc.abstractmethod
    def load(cls, path: str, **kwargs) -> "TraceDataset": ...

    @classmethod
    @abc.abstractmethod
    def new(
        cls,
        path: str,
        channel_names: list[str],
        trace_count: int,
        sample_count: int,
        data_length: int,
        version,
        **kwargs,
    ) -> "TraceDataset": ...

    @abc.abstractmethod
    def dump(self, path: str | None = None, **kwargs): ...

    @abc.abstractmethod
    def set_trace(self, channel_name: str, trace_index: int, trace: np.ndarray, data: np.ndarray | None): ...

    @property
    def trace_data(self) -> TraceDatasetData:
        return TraceDatasetData(get_trace_data=self._get_trace_data)

    @property
    def trace(self):
        return TraceDatasetData(get_trace_data=self._get_trace)

    @property
    def plain_text(self):
        return TraceDatasetData(get_trace_data=self._get_plaintext)

    @property
    def trace_data_with_indices(self):
        return TraceDatasetData(get_trace_data=self._get_trace_data_with_indices)

    def __getitem__(self, item):
        return TraceDatasetData(get_trace_data=self._get_trace_data_with_indices)[item]

    @abc.abstractmethod
    def _get_trace_data_with_indices(self, channel_slice, trace_slice) -> tuple[list, list, np.ndarray, np.ndarray]: ...

    # This method should return trace, plaintext, and ciphertext.
    # However, the ciphertext is currently empty, so it only returns trace and plaintext.
    @abc.abstractmethod
    def _get_trace_data(self, channel_slice, trace_slice) -> tuple[np.ndarray, np.ndarray]: ...

    @abc.abstractmethod
    def _get_trace(self, channel_slice, trace_slice) -> np.ndarray: ...

    @abc.abstractmethod
    def _get_plaintext(self, channel_slice, trace_slice) -> np.ndarray: ...

    @abc.abstractmethod
    def _get_ciphertext(self, channel_slice, trace_slice) -> np.ndarray: ...

    @staticmethod
    def _parse_slice(origin_count, index_slice) -> list:
        if origin_count is None:
            raise Exception("origin_count is not set")
        if isinstance(index_slice, slice):
            start, stop, step = index_slice.indices(origin_count)
            indices = [i for i in range(start, stop, step)]
        elif isinstance(index_slice, int):
            indices = [index_slice]
        elif isinstance(index_slice, list):
            indices = index_slice
        else:
            raise ValueError("index_slice is not a slice or list")
        return indices

    def __repr__(self):
        t = type(self)
        return f"<{t.__module__}.{t.__name__} ({self._channel_names}, {self._trace_count})"

    def info(self):
        return _InfoRender(
            self._channel_names, self._channel_count, self._trace_count, self._sample_count, self._data_length
        )

    @property
    def channel_names(self):
        return self._channel_names

    @property
    def channel_count(self):
        return self._channel_count

    @property
    def trace_count(self):
        return self._trace_count

    @property
    def sample_count(self):
        return self._sample_count

    @property
    def data_length(self):
        return self._data_length

    @property
    def create_time(self):
        return self._create_time


class _InfoRender:
    def __init__(
        self, channel_names: list[str], channel_count: int, trace_count: int, sample_count: int, data_length: int
    ):
        self._channel_names: list[str] = channel_names
        self._channel_count: int = channel_count
        self._trace_count: int = trace_count
        self._sample_count: int = sample_count
        self._data_length: int = data_length

    def __repr__(self):
        return (
            f"Channel: {self._channel_names}\r\n"
            f"Trace:   {self._trace_count}, {self._sample_count}\r\n"
            f"Data:    {self._trace_count}, {self._data_length}"
        )


class ScarrTraceDataset(TraceDataset):
    _ATTR_METADATA_KEY = "metadata"
    _GROUP_ROOT_PATH = "0"
    _ARRAY_TRACES_PATH = "traces"
    _ARRAY_PLAINTEXT_PATH = "plaintext"

    def __init__(
        self,
        zarr_path: str,
        create_empty: bool = False,
        channel_names: list[str] | None = None,
        trace_count: int | None = None,
        sample_count: int | None = None,
        data_length: int | None = None,
        trace_dtype: np.dtype = np.int16,
        zarr_kwargs: dict | None = None,
        zarr_trace_group_kwargs: dict | None = None,
        zarr_data_group_kwargs: dict | None = None,
        create_time: int | None = None,
        version: str | None = None,
    ):
        self._zarr_path: str = zarr_path
        self._channel_names: list[str] | None = channel_names
        self._channel_count = None if self._channel_names is None else len(self._channel_names)
        self._trace_count: int | None = trace_count
        self._sample_count: int | None = sample_count
        self._data_length: int | None = data_length
        self._create_time: int | None = create_time
        self._version: str | None = version

        if zarr_kwargs is None:
            zarr_kwargs = {}
        if zarr_trace_group_kwargs is None:
            zarr_trace_group_kwargs = {}
        if zarr_data_group_kwargs is None:
            zarr_data_group_kwargs = {}

        mode = zarr_kwargs.pop("mode", "w" if create_empty else "r")
        self._zarr_data = zarr.open(zarr_path, mode=mode, **zarr_kwargs)

        if create_empty:
            if (
                self._channel_names is None
                or self._trace_count is None
                or self._sample_count is None
                or self._data_length is None
            ):
                raise ValueError(
                    "channel_names and trace_count and sample_count and data_length "
                    "must be specified when in write mode."
                )
            self._create_time = int(time.time())
            group_root = self._zarr_data.create_group(self._GROUP_ROOT_PATH)
            for i, _ in enumerate(self._channel_names):
                channel_group = group_root.create_group(str(i))
                channel_group.create(
                    self._ARRAY_TRACES_PATH,
                    shape=(self._trace_count, self._sample_count),
                    dtype=trace_dtype,
                    **zarr_trace_group_kwargs,
                )
                channel_group.create(
                    self._ARRAY_PLAINTEXT_PATH,
                    shape=(self._trace_count, self._data_length),
                    dtype=np.uint8,
                    **zarr_data_group_kwargs,
                )
            self._zarr_data.attrs[self._ATTR_METADATA_KEY] = {
                "create_time": self._create_time,
                "channel_names": self._channel_names,
                "trace_count": self._trace_count,
                "sample_count": self._sample_count,
                "data_length": self._data_length,
                "version": self._version,
            }
        else:
            if self._zarr_path is None:
                raise ValueError("The zarr_path must be specified when in non-write mode.")
            metadata = self._zarr_data.attrs[self._ATTR_METADATA_KEY]
            self._create_time = metadata.get("create_time")
            # This is a piece of logic for handling dataset files compatible with previous versions,
            # which will be removed in subsequent stable versions.
            if "channel_names" not in metadata:
                self._channel_count = metadata.get("channel_count")
                self._channel_names = [str(i) for i in range(self._channel_count)]
            else:
                self._channel_names = metadata.get("channel_names")
                self._channel_count = len(self._channel_names)
            self._trace_count = metadata.get("trace_count")
            self._sample_count = metadata.get("sample_count")
            self._data_length = metadata.get("data_length")
            self._version = metadata.get("version")

    @classmethod
    def load(cls, path: str, **kwargs) -> "TraceDataset":
        kwargs["mode"] = "r"
        return cls(path, zarr_kwargs=kwargs)

    @classmethod
    def new(
        cls,
        path: str,
        channel_names: list[str],
        trace_count: int,
        sample_count: int,
        data_length: int,
        version: str,
        **kwargs,
    ) -> "TraceDataset":
        kwargs["mode"] = "w"
        return cls(
            path,
            create_empty=True,
            channel_names=channel_names,
            trace_count=trace_count,
            sample_count=sample_count,
            data_length=data_length,
            version=version,
            zarr_kwargs=kwargs,
        )

    def dump(self, path: str | None = None, **kwargs):
        if path is not None and path != self._zarr_path:
            zarr.copy_store(self._zarr_data, zarr.open(path, mode="w"))

    def set_trace(self, channel_name: str, trace_index: int, trace: np.ndarray, data: np.ndarray | None):
        if self._trace_count is None or self._channel_count is None:
            raise Exception("Channel or trace count must has not specified.")
        if channel_name not in self._channel_names:
            raise ValueError("channel index out range")
        if trace_index not in range(0, self._trace_count):
            raise ValueError("trace, index out of range")
        channel_index = self._channel_names.index(channel_name)
        self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_index] = trace
        if self._data_length != 0 and data is not None:
            self._get_under_root(channel_index, self._ARRAY_PLAINTEXT_PATH)[trace_index] = data

    def get_origin_data(self) -> zarr.hierarchy.Group:
        return self._zarr_data

    def get_trace_by_indexes(self, channel_name: str, *trace_indexes: int) -> tuple[np.ndarray, np.ndarray] | None:
        channel_index = self._channel_names.index(channel_name)
        return (
            self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[[i for i in trace_indexes]],
            self._get_under_root(channel_index, self._ARRAY_PLAINTEXT_PATH)[[i for i in trace_indexes]],
        )

    def get_trace_by_range(
        self, channel_name: str, index_start: int, index_end: int
    ) -> tuple[np.ndarray, np.ndarray] | None:
        channel_index = self._channel_names.index(channel_name)
        return (
            self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[index_start:index_end],
            self._get_under_root(channel_index, self._ARRAY_PLAINTEXT_PATH)[index_start:index_end],
        )

    def _get_under_root(self, *paths: typing.Any):
        paths = self._GROUP_ROOT_PATH, *paths
        return self._zarr_data["/".join(str(path) for path in paths)]

    def _get_trace_data_with_indices(self, channel_slice, trace_slice) -> tuple[list, list, np.ndarray, np.ndarray]:
        traces = []
        data = []

        channel_indexes, trace_indexes = (
            self._parse_slice(self._channel_count, channel_slice),
            self._parse_slice(self._trace_count, trace_slice),
        )

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        for channel_index in channel_indexes:
            traces.append(self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_slice])
            data.append(self._get_under_root(channel_index, self._ARRAY_PLAINTEXT_PATH)[trace_slice])

        return channel_indexes, trace_indexes, np.array(traces), np.array(data)

    def _get_trace_data(self, channel_slice, trace_slice) -> tuple[np.ndarray, np.ndarray]:
        traces = []
        plaintext = []

        channel_indexes = self._parse_slice(self.channel_count, channel_slice)

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        for channel_index in channel_indexes:
            traces.append(self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_slice])
            plaintext.append(self._get_under_root(channel_index, self._ARRAY_PLAINTEXT_PATH)[trace_slice])

        return np.vstack(traces), np.vstack(plaintext)

    def _get_trace(self, channel_slice, trace_slice) -> np.ndarray:
        traces = []

        channel_indexes = self._parse_slice(self.channel_count, channel_slice)

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        for channel_index in channel_indexes:
            traces.append(self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_slice])

        return np.vstack(traces) if len(traces) == 1 else np.stack(traces)

    def _get_plaintext(self, channel_slice, trace_slice) -> np.ndarray:
        plaintext = []

        channel_indexes = self._parse_slice(self.channel_count, channel_slice)

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        for channel_index in channel_indexes:
            plaintext.append(self._get_under_root(channel_index, self._ARRAY_PLAINTEXT_PATH)[trace_slice])

        return np.vstack(plaintext) if len(plaintext) == 1 else np.stack(plaintext)

    def _get_ciphertext(self, channel_slice, trace_slice) -> np.ndarray: ...


class NumpyTraceDataset(TraceDataset):
    _ARRAY_TRACE_PATH = "trace.npy"
    _ARRAY_DATA_PATH = "data.npy"
    _METADATA_PATH = "metadata.json"

    def __init__(
        self,
        npy_trace_path: str | None = None,
        npy_data_path: str | None = None,
        npy_metadata_path: str | None = None,
        create_empty: bool = False,
        channel_names: list[str] | None = None,
        trace_count: int | None = None,
        sample_count: int | None = None,
        trace_dtype: np.dtype = np.int16,
        data_length: int | None = None,
        create_time: int | None = None,
        version: str | None = None,
    ):
        self._logger = logger.get_logger(NumpyTraceDataset)

        self._npy_trace_path: str | None = npy_trace_path
        self._npy_data_path: str | None = npy_data_path
        self._npy_metadata_path: str | None = npy_metadata_path

        self._channel_names: list[str] | None = channel_names
        self._channel_count: int | None = None if self._channel_names is None else len(self._channel_names)
        self._trace_count: int | None = trace_count
        self._sample_count: int | None = sample_count
        self._data_length: int | None = data_length
        self._create_time: int | None = create_time
        self._version: str | None = version

        self._trace_array: np.ndarray
        self._data_array: np.ndarray

        if create_empty:
            if (
                self._channel_names is None
                or self._trace_count is None
                or self._sample_count is None
                or self._data_length is None
            ):
                raise ValueError(
                    "channel_names and trace_count and sample_count and data_length "
                    "must be specified when in write mode."
                )
            self._trace_array = np.zeros(
                shape=(self._channel_count, self._trace_count, self._sample_count), dtype=trace_dtype
            )
            self._plaintext_array = np.zeros(
                shape=(self._channel_count, self._trace_count, self._data_length), dtype=np.uint8
            )
            self._create_time = int(time.time())

        else:
            if self._npy_trace_path is None:
                raise ValueError("The npy_trace_path must be specified when in non-write mode.")

            self._trace_array = np.load(self._npy_trace_path)

            if self._npy_data_path is None or self._npy_metadata_path is None:
                self._logger.warning(
                    "npy_data_path or npy_metadata_path is not specified, data or metadata info will be not load."
                )
            else:
                self._plaintext_array = np.load(self._npy_data_path)
                self._load_metadata()

    def _load_metadata(self):
        with open(self._npy_metadata_path) as f:
            metadata = json.load(f)
            # This is a piece of logic for handling dataset files compatible with previous versions,
            # which will be removed in subsequent stable versions.
            if "channel_names" not in metadata:
                self._channel_count = metadata["channel_count"]
                self._channel_names = [str(i) for i in range(self._channel_count)]
            else:
                self._channel_names: list[str] | None = metadata.get("channel_names")
                self._channel_count: int | None = len(self._channel_names)
            self._trace_count: int | None = metadata.get("trace_count")
            self._sample_count: int | None = metadata.get("sample_count")
            self._data_length: int | None = metadata.get("data_length")
            self._create_time: int | None = metadata.get("create_time")
            self._version: str | None = metadata.get("version")

    def _dump_metadata(self):
        with open(self._npy_metadata_path, "w") as f:
            json.dump(
                {
                    "channel_names": self._channel_names,
                    "trace_count": self._trace_count,
                    "sample_count": self._sample_count,
                    "data_length": self._data_length,
                    "create_time": self._create_time,
                    "version": self._version,
                },
                f,
            )

    def get_origin_data(self) -> tuple[np.array, np.array]:
        return self._trace_array, self._plaintext_array

    @classmethod
    def load(cls, path: str, **kwargs) -> "TraceDataset":
        return cls(
            os.path.join(path, cls._ARRAY_TRACE_PATH),
            os.path.join(path, cls._ARRAY_DATA_PATH),
            os.path.join(path, cls._METADATA_PATH),
            **kwargs,
        )

    @classmethod
    def load_from_numpy_array(cls, trace: np.ndarray, data: np.ndarray | None = None):
        channel_count = None
        trace_count = None
        sample_count = None
        data_length = None

        shape = trace.shape

        if data is not None and not shape == data.shape:
            raise ValueError("trace and data must have the same shape.")

        array_size = len(shape)

        if array_size == 1:
            channel_count = 1
            trace_count = 1
            sample_count = shape[0]
            data_length = data.shape[0] if data is not None else 0
        elif array_size == 2:
            channel_count = 1
            trace_count = shape[0]
            sample_count = shape[1]
            data_length = data.shape[1] if data is not None else 0
        elif array_size == 3:
            channel_count = shape[0]
            trace_count = shape[1]
            sample_count = shape[2]
            data_length = data.shape[2] if data is not None else 0

        channel_names = []
        for i in range(channel_count):
            channel_names.append(str(i))

        ds = cls(
            create_empty=True,
            channel_names=channel_names,
            trace_count=trace_count,
            sample_count=sample_count,
            data_length=data_length,
            trace_dtype=trace.dtype,
        )

        if array_size == 1:
            ds.set_trace(0, 0, trace, data)
        elif array_size == 2:
            for t in range(shape[0]):
                ds.set_trace(0, t, trace[t], data[t] if data is not None else None)
        elif array_size == 3:
            for c in range(shape[0]):
                for t in range(shape[1]):
                    ds.set_trace(c, t, trace[c, t], data[c, t] if data is not None else None)

        return ds

    @classmethod
    def new(
        cls,
        path: str,
        channel_names: list[str],
        trace_count: int,
        sample_count: int,
        data_length: int,
        version: str,
        **kwargs,
    ) -> "TraceDataset":
        if not os.path.exists(path):
            os.makedirs(path)
        elif os.path.isfile(path):
            raise Exception(f"{path} is not a file.")

        npy_trace_path = os.path.join(path, cls._ARRAY_TRACE_PATH)
        npy_data_path = os.path.join(path, cls._ARRAY_DATA_PATH)
        npy_metadata_path = os.path.join(path, cls._METADATA_PATH)

        return cls(
            npy_trace_path=npy_trace_path,
            npy_data_path=npy_data_path,
            npy_metadata_path=npy_metadata_path,
            create_empty=True,
            channel_names=channel_names,
            trace_count=trace_count,
            sample_count=sample_count,
            data_length=data_length,
            version=version,
            **kwargs,
        )

    def dump(self, path: str | None = None, **kwargs):
        if self._npy_trace_path is None or self._npy_data_path is None:
            raise Exception("trace and metadata path must not be None.")
        else:
            np.save(self._npy_trace_path, self._trace_array)
            np.save(self._npy_data_path, self._plaintext_array)
            self._dump_metadata()

    def set_trace(self, channel_name: str | int, trace_index: int, trace: np.ndarray, data: np.ndarray | None):
        if isinstance(channel_name, int):
            channel_index = channel_name
        else:
            channel_index = self._channel_names.index(channel_name)
        self._trace_array[channel_index, trace_index, :] = trace
        if self._data_length != 0 and data is not None:
            self._plaintext_array[channel_index, trace_index, :] = data

    def _get_trace_data_with_indices(self, channel_slice, trace_slice) -> tuple[list, list, np.ndarray, np.ndarray]:
        c, t = self._parse_slice(channel_slice, trace_slice)
        if isinstance(channel_slice, int):
            channel_slice = slice(channel_slice, channel_slice + 1)
        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)
        return c, t, self._trace_array[channel_slice, trace_slice], self._plaintext_array[channel_slice, trace_slice]

    def _get_trace_data(self, channel_slice, trace_slice) -> tuple[np.ndarray, np.ndarray]:
        pass

    def _get_trace(self, channel_slice, trace_slice) -> np.ndarray:
        return self._trace_array[channel_slice, trace_slice]

    def _get_plaintext(self, channel_slice, trace_slice) -> np.ndarray:
        return self._plaintext_array[channel_slice, trace_slice]

    def _get_ciphertext(self, channel_slice, trace_slice) -> np.ndarray:
        pass



# Copyright 2024 CrackNuts. All rights reserved.

import typing

from anywidget import AnyWidget


class MsgHandlerPanelWidget(AnyWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(**{})
        self._msg_handlers: dict[str, list[typing.Callable[[dict[str, typing.Any]], None]]] = {}
        self.on_msg(self._msg_handle)

    def _msg_handle(self, widget, content: dict[str, typing.Any], buffers):
        source = content.get("source")
        event = content.get("event")
        args: dict = content.get("args")
        key = self._get_msg_handler_key(source, event)
        if key in self._msg_handlers:
            for handler in self._msg_handlers[key]:
                handler(args)

    @staticmethod
    def _get_msg_handler_key(source, event):
        return source + "_" + event

    def reg_msg_handler(self, source, event, handler: typing.Callable[[dict[str, typing.Any]], None]):
        key = self._get_msg_handler_key(source, event)
        if key not in self._msg_handlers:
            self._msg_handlers[key] = []
        self._msg_handlers[key].append(handler)



# Copyright 2024 CrackNuts. All rights reserved.

import logging
import threading
import time

from cracknuts.cracker import protocol
from cracknuts.mock.mock_cracker import MockCracker
from cracknuts.mock.mock_operator import MockOperator


def start(
    host: str = "127.0.0.1",
    port: int = protocol.DEFAULT_PORT,
    operator_port: int = protocol.DEFAULT_OPERATOR_PORT,
    logging_level: str | int = logging.INFO,
):
    cracker = MockCracker(host, port, logging_level)
    cracker_thread = threading.Thread(target=cracker.start)
    cracker_thread.start()

    operator = MockOperator(host, operator_port, logging_level)
    operator_thread = threading.Thread(target=operator.start)
    operator_thread.start()

    print(f"Mock cracker start success, cracker server listen on {port} and operator listen on {operator_port}.")

    try:
        while cracker_thread.is_alive() or operator_thread.is_alive():
            time.sleep(0.3)
    except KeyboardInterrupt:
        cracker.stop()
        operator.stop()


if __name__ == "__main__":
    start()



# Copyright 2024 CrackNuts. All rights reserved.

import logging
import os
import socket
import struct
import sys
import threading
import time

import numpy as np

from cracknuts import logger
from cracknuts.cracker import protocol
from cracknuts.utils import hex_util

_handler_dict = {}


def _handler(command: int, has_payload: bool = True):
    def decorator(func):
        def wrapper(*args, **kwargs):
            if not has_payload:
                del kwargs["payload"]
            return func(*args, **kwargs)

        _handler_dict[command] = wrapper
        return wrapper

    return decorator


class MockCracker:
    def __init__(self, host: str = "127.0.0.1", port: int = protocol.DEFAULT_PORT, logging_level=logging.INFO):
        self._running = False
        self._logger = logger.get_logger(MockCracker, logging_level)
        self._server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server_socket.settimeout(1)
        self._server_socket.bind((host, port))
        self._server_socket.listen()
        self._logger.debug("MockCracker initialized")
        self._command_payload_cache = {}

    def start(self):
        try:
            self._running = True
            self._start_accept()
        except KeyboardInterrupt:
            self._logger.debug("exist by keyboard interrupt.")
            sys.exit(0)

    def stop(self):
        self._running = False

    def _start_accept(self):
        while self._running:
            try:
                conn, addr = self._server_socket.accept()
                conn.settimeout(5)
                self._logger.debug(f"Get client request from {addr}.")
                threading.Thread(target=self._handle, args=(conn,), daemon=True).start()
                # self._handle(conn)
            except TimeoutError:
                continue

    def _handle(self, conn):
        while self._running:
            try:
                header_data = conn.recv(protocol.REQ_HEADER_SIZE)
                if not header_data:
                    self._logger.debug("disconnected...")
                    break

                self._logger.debug(f"Received header:\n{hex_util.get_bytes_matrix(header_data)}")

                try:
                    magic, version, direction, command, rfu, length = struct.unpack(
                        protocol.REQ_HEADER_FORMAT, header_data
                    )
                except struct.error as e:
                    err_msg = f"Message format error: {e.args[0]}\n"
                    self._logger.error(err_msg)
                    err_res = self._error(f"Header format error: {hex_util.get_hex(header_data)}")
                    self._logger.debug(f"Send error:\n{hex_util.get_bytes_matrix(err_res)}")
                    conn.sendall(err_res)
                    continue

                self._logger.info(f"Received command: 0x{command:02X}")
                self._logger.debug(
                    f"Received header: Magic={magic}, Version={version}, Direction={direction}, "
                    f"Command={command}, Length={length}"
                )

                payload_data = conn.recv(length)
                self._command_payload_cache[command] = payload_data

                if len(payload_data) > 0:
                    if self._logger.isEnabledFor(logging.INFO):
                        self._logger.info(f"Received payload: {hex_util.get_hex(payload_data)}")
                    if self._logger.isEnabledFor(logging.DEBUG):
                        self._logger.debug(f"Received payload:\n{hex_util.get_bytes_matrix(payload_data)}")
                if command not in _handler_dict:
                    self._logger.warning(f'Get command not supported: 0x{format(command, '04x')}')
                    unsupported_res = self._unsupported(command)
                    self._logger.debug(f"Send unsupported:\n{hex_util.get_bytes_matrix(unsupported_res)}")
                    conn.sendall(unsupported_res)
                    continue
                func_res = _handler_dict.get(command, None)(self, payload=payload_data)
                if type(func_res) is tuple:
                    status, res_payload = func_res
                else:
                    status, res_payload = protocol.STATUS_OK, func_res
                res = protocol.build_response_message(status, res_payload)
                self._logger.debug(f"Sending:\n{hex_util.get_bytes_matrix(res)}")
                conn.sendall(res)
            except TimeoutError:
                continue
            except ConnectionResetError:
                break
            except Exception as e:
                self._logger.error(e)
                raise e

    @staticmethod
    def _error(error: str) -> bytes:
        return protocol.build_response_message(protocol.STATUS_ERROR, error.encode())

    @staticmethod
    def _unsupported(command):
        return protocol.build_response_message(
            protocol.STATUS_OK, f'Command [0x{format(command, '04x')}] not supported'.encode()
        )

    @_handler(protocol.Command.GET_ID, has_payload=False)
    def get_id(self) -> bytes:
        return b"mock_001"

    @_handler(protocol.Command.GET_NAME, has_payload=False)
    def get_name(self) -> bytes:
        return b"MOCK 001"

    @_handler(protocol.Command.GET_VERSION, has_payload=False)
    def get_version(self) -> bytes:
        return b"0.1.0(adc:0.1.0,hardware:0.1.0)"

    @_handler(protocol.Command.OSC_GET_ANALOG_WAVES)
    def osc_get_analog_wave(self, payload: bytes) -> bytes:
        channel, offset, sample_count = struct.unpack(">BII", payload)
        if channel == 1:
            mn, mx = -26, -1
        elif channel == 2:
            mn, mx = 0, 25
        else:
            mn, mx = -1, 1
        # mn, mx = -50, 50
        return struct.pack(f"<{sample_count}h", *np.random.randint(mn, mx, size=sample_count).tolist())

    @_handler(protocol.Command.OSC_IS_TRIGGERED)
    def osc_is_trigger(self, payload: bytes) -> tuple[int, bytes | None]:
        time.sleep(0.05)  # Simulate device I/O operations.
        return protocol.STATUS_OK, int.to_bytes(4, 4, "big")

    @_handler(protocol.Command.OSC_FORCE, has_payload=False)
    def osc_force(self) -> bytes: ...  # type: ignore

    @_handler(protocol.Command.OSC_SINGLE, has_payload=False)
    def osc_arm(self) -> bytes: ...  # type: ignore

    @_handler(protocol.Command.NUT_VOLTAGE, has_payload=False)
    def nut_voltage(self) -> bytes: ...  # type: ignore

    @_handler(protocol.Command.OSC_ANALOG_CHANNEL_ENABLE, has_payload=False)
    def osc_analog_channel_enable(self) -> bytes: ...  # type: ignore

    @_handler(protocol.Command.NUT_CLOCK, has_payload=False)
    def nut_clock(self) -> bytes: ...  # type: ignore

    @_handler(protocol.Command.OSC_SAMPLE_LENGTH, has_payload=False)
    def osc_sample_len(self) -> bytes: ...

    @_handler(protocol.Command.NUT_ENABLE, has_payload=False)
    def nut_enable(self) -> bytes: ...

    @_handler(protocol.Command.OSC_ANALOG_GAIN, has_payload=False)
    def osc_analog_gain(self) -> bytes: ...

    @_handler(protocol.Command.OSC_SAMPLE_DELAY, has_payload=False)
    def osc_sample_delay(self) -> bytes: ...

    @_handler(protocol.Command.OSC_SAMPLE_RATE, has_payload=False)
    def osc_sample_rate(self) -> bytes: ...

    @_handler(protocol.Command.OSC_CLOCK_SAMPLE_PHASE, has_payload=False)
    def osc_clock_sample_phase(self) -> bytes: ...

    @_handler(protocol.Command.OSC_ANALOG_TRIGGER_SOURCE, has_payload=False)
    def osc_analog_trigger_source(self) -> bytes: ...

    @_handler(protocol.Command.OSC_TRIGGER_MODE, has_payload=False)
    def osc_trigger_mode(self) -> bytes: ...

    @_handler(protocol.Command.OSC_TRIGGER_EDGE, has_payload=False)
    def osc_trigger_edge(self) -> bytes: ...

    @_handler(protocol.Command.OSC_TRIGGER_EDGE_LEVEL, has_payload=False)
    def osc_trigger_edge_level(self) -> bytes: ...

    @_handler(protocol.Command.CRACKER_SERIAL_DATA)
    def cracker_serial_data(self, payload: bytes) -> bytes:
        return os.urandom(10)

    @_handler(protocol.Command.NUT_CLOCK_ENABLE)
    def nut_clock_enable(self, payload: bytes) -> bytes:
        return b""

    @_handler(0xFFFF)
    def get_command_payload_cache(self, payload: bytes) -> bytes:
        command = struct.unpack(">I", payload)[0]
        return self._command_payload_cache.get(command)



# Copyright 2024 CrackNuts. All rights reserved.

import pathlib
import typing

import numpy as np
from cracknuts.trace.trace import TraceDataset, NumpyTraceDataset
from traitlets import traitlets

from cracknuts.jupyter.panel import MsgHandlerPanelWidget


class TracePanelWidget(MsgHandlerPanelWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "TracePanelWidget.js"
    _css = ""

    trace_series_list: list[dict] = traitlets.List([])

    _DEFAULT_SHOW_INDEX_THRESHOLD = 3

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(*args, **kwargs)
        self._emphasis = kwargs.get("emphasis", False)
        self._trace_dataset: TraceDataset | None = None
        self.show_trace: ShowTrace = ShowTrace(self._show_trace)
        self._auto_sync = False

    def _repr_mimebundle_(self, **kwargs: dict) -> tuple[dict, dict] | None:
        self.send_state("trace_series_list")
        self._auto_sync = True
        return super()._repr_mimebundle_(**kwargs)

    def set_emphasis(self, value: bool) -> None:
        if self._emphasis != value:
            self._emphasis = value
            self.trace_series_list = [{**series, "emphasis": not value} for series in self.trace_series_list]
            if self._auto_sync:
                self.send_state("trace_series_list")

    def set_numpy_data(self, trace: np.ndarray, data: np.ndarray = None) -> None:
        ds = NumpyTraceDataset.load_from_numpy_array(trace, data)
        self.set_trace_dataset(ds)

    def _show_trace(self, channel_slice, trace_slice):
        channel_indexes, trace_indexes, traces, data = self._trace_dataset.trace_data_with_indices[
            channel_slice, trace_slice
        ]

        trace_series_list = []

        for c, channel_index in enumerate(channel_indexes):
            for t, trace_index in enumerate(trace_indexes):
                trace_series_list.append(
                    {
                        "name": str(channel_index) + "-" + str(trace_index),
                        "data": traces[c, t],
                        "emphasis": not self._emphasis,
                    }
                )
        self.trace_series_list = trace_series_list
        if self._auto_sync:
            self.send_state("trace_series_list")

    def show_default_trace(self):
        trace_count = self._trace_dataset.trace_count
        self._show_trace(
            0,
            slice(
                0,
                trace_count if trace_count < self._DEFAULT_SHOW_INDEX_THRESHOLD else self._DEFAULT_SHOW_INDEX_THRESHOLD,
            ),
        )

    def show_all_trace(self):
        self._show_trace(slice(0, self._trace_dataset.channel_count), slice(0, self._trace_dataset.trace_count))

    def highlight(self, *indexes: int) -> None:
        self._emphasis = False
        x = [
            {**series, "emphasis": True, "color": "red" if i in indexes else "gray", "z": 100 if i in indexes else 1}
            for i, series in enumerate(self.trace_series_list)
        ]
        self.trace_series_list = x
        if self._auto_sync:
            self.send_state("trace_series_list")

    def set_numpy_data_highlight(self, trace: np.ndarray, data: np.ndarray = None, highlight_indexes=None):
        if highlight_indexes is None:
            highlight_indexes = []
        ds = NumpyTraceDataset.load_from_numpy_array(trace, data)
        self._trace_dataset = ds
        trace_series_list = []

        for t, trace_index in enumerate(range(ds.trace_count)):
            if t not in highlight_indexes:
                trace_series_list.append(
                    {
                        "name": str(0) + "-" + str(trace_index),
                        "data": trace[t],
                        "emphasis": not self._emphasis,
                        "color": "gray",
                    }
                )
        for t, trace_index in enumerate(range(ds.trace_count)):
            if t in highlight_indexes:
                trace_series_list.append(
                    {
                        "name": str(0) + "-" + str(trace_index),
                        "data": trace[t],
                        "emphasis": not self._emphasis,
                        "color": "red",
                        "z": 100,
                    }
                )
        self.trace_series_list = trace_series_list
        if self._auto_sync:
            self.send_state("trace_series_list")

    def set_trace_dataset(
        self, trace_dataset: TraceDataset, show_all_trace=False, channel_slice=None, trace_slice=None
    ) -> None:
        self._trace_dataset = trace_dataset
        if show_all_trace:
            self.show_all_trace()
        elif channel_slice is not None and trace_slice is not None:
            self._show_trace(channel_slice, trace_slice)
        else:
            self.show_default_trace()

    def open_trace_data_set_file(self, trace_dataset: TraceDataset) -> TraceDataset: ...

    def open_numpy_file(self, path: str): ...

    def open_zarr(self, path: str): ...


class ShowTrace:
    def __init__(self, func_show_trace):
        self._func_trace = func_show_trace

    def __getitem__(self, item):
        return self._func_trace(item[0], item[1])



# Copyright 2024 CrackNuts. All rights reserved.

import threading
import time
import typing

import numpy as np

from cracknuts.cracker.cracker_basic import CrackerBasic
from cracknuts import logger
from cracknuts.cracker import protocol


class ScopeAcquisition:
    def __init__(self, cracker: CrackerBasic, interval: float = 0.05, repeat_interval: float = 0.1):
        self._logger = logger.get_logger(self)
        self._status: int = 0  # 0 stop, 1 normal, 2 single, 3 repeat
        self._cracker: CrackerBasic = cracker
        self._offset: int = 0
        self._last_waves = {}
        self._run_thread_pause_event: threading.Event = threading.Event()
        self._interval: float = interval
        self._trigger_judge_wait_time: float = 0.01
        self._repeat_interval: float = repeat_interval
        self._status_change_listener: list[typing.Callable[[int], None]] = []

    def on_status_changed(self, callback: typing.Callable[[int], None]) -> None:
        self._status_change_listener.append(callback)

    def _status_changed(self):
        for listener in self._status_change_listener:
            listener(self._status)

    def set_offset(self, offset: int):
        self._offset = offset

    def is_running(self) -> bool:
        return self._status != 0

    def run(self, model: int = 1) -> None:
        if model == 0:
            self.stop()
        else:
            if self._status == 0:
                # When starting from the stopped state, clear the waveform data left from the previous acquisition.
                self._last_waves = {}
                self._status = model
                threading.Thread(target=self._acquisition, kwargs={"model": model}).start()
                self._status_changed()
            else:
                self._status = model

    def stop(self):
        if not self._run_thread_pause_event.is_set():
            self._run_thread_pause_event.set()
        self._status = 0
        self._status_changed()

    def resume(self):
        if self._status < 0:
            self._run_thread_pause_event.clear()
            self._status = -1 * abs(self._status)
            self._status_changed()

    def pause(self):
        self._status = abs(self._status)
        self._status_changed()

    def start_normal(self):
        self.run(1)

    def start_single(self):
        self.run(2)

    def start_repeat(self):
        self.run(3)

    def _acquisition(self, model: int):
        while self._status != 0:
            if self._status < 0:
                if self._status == -2:
                    # Pause is not supported in single model. it will exit.
                    self._run_thread_pause_event.set()
                    self.stop()
                else:
                    self._run_thread_pause_event.wait()
            elif self._status == 1:
                self._cracker.osc_force()
                self._last_waves = self._get_waves()
                time.sleep(self._interval)
            elif self._status == 2:
                while self._status == 2:
                    self._cracker.osc_single()
                    if self._cracker.osc_is_triggered():
                        self._last_waves = self._get_waves()
                        self.stop()
                    time.sleep(self._trigger_judge_wait_time)
            elif self._status == 3:
                self._cracker.osc_single()
                trigger_jude_start_time = time.time()
                while self._status == 3:
                    if time.time() - trigger_jude_start_time > self._repeat_interval:
                        self._cracker.osc_force()
                    else:
                        if self._cracker.osc_is_triggered():
                            self._last_waves = self._get_waves()
                            break
                        time.sleep(self._trigger_judge_wait_time)
                self._last_waves = self._get_waves()
            else:
                self._logger.error(f"scope_acquisition error: {self._status}")

    def _get_waves(self):
        config = self._cracker.get_current_config()
        if config.osc_analog_channel_enable is None:
            raise Exception("Channel info can't be none.")
        enable_channels = [k for k, v in config.osc_analog_channel_enable.items() if v]
        wave_dict = {}
        for c in enable_channels:
            status, wave = self._cracker.osc_get_analog_wave(
                c, self._offset, self._cracker.get_current_config().osc_sample_length
            )
            if status == protocol.STATUS_OK:
                wave_dict[c] = wave
            else:
                del wave_dict[c]
        return wave_dict

    def get_last_wave(self) -> dict[int, np.ndarray]:
        return self._last_waves



# Copyright 2024 CrackNuts. All rights reserved.

import abc
import datetime
import json
import os.path
import threading
import time
import traceback
import typing

import numpy as np

from cracknuts import logger
from cracknuts.cracker.cracker_basic import CrackerBasic
from cracknuts.trace.trace import ScarrTraceDataset, NumpyTraceDataset


class AcqProgress:
    def __init__(self, finished: int, total: int):
        self.finished: int = finished
        self.total: int = total


class Acquisition(abc.ABC):
    """
    Trace acquisition process control class
    """

    STATUS_STOPPED: int = 0
    STATUS_TESTING: int = 1
    STATUS_RUNNING: int = 2
    _STATUS_PAUSED_SWITCH: int = -1

    DO_ERROR_HANDLER_STRATEGY_EXIT: int = 0
    DO_ERROR_HANDLER_STRATEGY_CONTINUE_UNTIL_MAX_ERROR_COUNT: int = 1

    _DATASET_DEFAULT_PATH = "./dataset/"

    def __init__(
        self,
        cracker: CrackerBasic,
        trace_count: int = 1000,
        sample_length: int = -1,
        sample_offset: int = 0,
        data_length: int = 0,
        trigger_judge_wait_time: float = 0.05,
        trigger_judge_timeout: float = 1.0,
        do_error_handler_strategy: int = DO_ERROR_HANDLER_STRATEGY_EXIT,
        do_error_max_count: int = -1,
        file_format: str = "scarr",
        file_path: str = "auto",
    ):
        """
        :param cracker: The controlled Cracker object.
        :type cracker: CrackerBasic
        :param trace_count: The number of traces to acquire.
        :type trace_count: int
        :param sample_length: the sample length to be acquired,
                              If it is -1, it will be set to the value of `cracker.current_count().sample_len`.
        :type sample_length: int
        :param sample_offset: the sample offset to be acquired.
        :type sample_offset: int
        :param data_length: the data length to be acquired.
        :type data_length: int
        :param trigger_judge_timeout: The trigger judge timeout in seconds.
        :type trigger_judge_timeout: float
        :param trigger_judge_wait_time: The trigger judge wait time in seconds.
        :type trigger_judge_wait_time: float
        :param do_error_handler_strategy: The strategy to handle error handling.
                                          0: Exit immediately, 1: Exit after exceeding the error count.
        :type do_error_handler_strategy: int
        :param do_error_max_count: The maximum number of error count.
                                   If `do_error_handler_strategy` is 0, this setting is invalid.
        :type do_error_max_count: int
        :param file_format: The file format of the trace dataset. "scarr" or "numpy".
        :type file_format: str
        :param file_path: The file path of the trace dataset. If set to "auto", a folder with a timestamp format
                          will be created in the current working directory to save the data.
        :type file_path: str
        """
        self._logger = logger.get_logger(self)
        self._last_wave: dict[int, np.ndarray] | None = {1: np.zeros(1)}
        self._status: int = self.STATUS_STOPPED
        self._current_trace_count = 1
        self._run_thread_pause_event: threading.Event = threading.Event()

        self.cracker: CrackerBasic = cracker
        self.trace_count: int = trace_count
        self.sample_length = sample_length
        self.sample_offset: int = sample_offset
        self.data_length: int = data_length
        self.trigger_judge_wait_time: float = trigger_judge_wait_time  # second
        self.trigger_judge_timeout: float = trigger_judge_timeout  # second
        self.do_error_handler_strategy: int = do_error_handler_strategy
        self.do_error_max_count: int = do_error_max_count  # -1 never exit
        self.file_format: str = file_format
        if file_path is None or file_path == "auto":
            file_path = os.path.abspath(self._DATASET_DEFAULT_PATH)
        self.file_path: str = file_path

        self._on_wave_loaded_callback: typing.Callable[[typing.Any], None] | None = None
        self._on_status_change_listeners: list[typing.Callable[[int], None]] = []
        self._on_run_progress_changed_listeners: list[typing.Callable[[dict], None]] = []

    def get_status(self):
        return self._status

    def set_cracker(self, cracker: CrackerBasic):
        self.cracker = cracker

    def on_status_changed(self, callback: typing.Callable[[int], None]) -> None:
        """
        User should not use this function. If you need to perform actions when the ACQ state changes,
        please use the node functions in the ACQ lifecycle point: `init`, `do`, and `finish`.

        status: 0 stopped, 1 testing, 2 running 3 paused
        """
        self._on_status_change_listeners.append(callback)

    def on_run_progress_changed(self, callback: typing.Callable[[dict], None]) -> None:
        self._on_run_progress_changed_listeners.append(callback)

    def on_wave_loaded(self, callback: typing.Callable[[typing.Any], None]) -> None:
        self._on_wave_loaded_callback = callback

    def run(
        self,
        count: int = 1,
        sample_length: int = 1024,
        sample_offset: int = 0,
        data_length: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        file_format: str | None = "scarr",
        file_path: str | None = "auto",
    ):
        """
        Start run mode in the background.
        The parameters configured here will override the settings specified in `__init__`.

        :param count: The number of traces to acquire.
        :type count: int
        :param sample_length: The sample length to be acquired.
        :type sample_length: int
        :param sample_offset: The sample offset to be acquired.
        :type sample_offset: int
        :param data_length: The data length to be acquired.
        :type data_length: int
        :param trigger_judge_wait_time: The trigger judge wait time in seconds.
        :type trigger_judge_wait_time: float
        :param trigger_judge_timeout: The trigger judge timeout in seconds.
        :type trigger_judge_timeout: float
        :param do_error_max_count: The maximum number of error count.
                                   If `do_error_handler_strategy` is 0, this setting is invalid.
        :type do_error_max_count: int
        :param do_error_handler_strategy: The strategy to handle error handling.
                                          0: Exit immediately, 1: Exit after exceeding the error count.
        :type do_error_handler_strategy: int
        :param file_format: The file format of the trace dataset. "scarr" or "numpy".
        :type file_format: str
        :param file_path: The file path of the trace dataset. If set to "auto", a folder with a timestamp format
                          will be created in the current working directory to save the data.
        :type file_path: str
        """
        if self._status < 0:
            self.resume()
        else:
            self._run(
                test=False,
                count=count,
                sample_length=sample_length,
                sample_offset=sample_offset,
                data_length=data_length,
                trigger_judge_wait_time=trigger_judge_wait_time,
                trigger_judge_timeout=trigger_judge_timeout,
                do_error_max_count=do_error_max_count,
                do_error_handler_strategy=do_error_handler_strategy,
                file_format=file_format,
                file_path=file_path,
            )

    def run_sync(
        self,
        count=1,
        sample_length: int = 1024,
        sample_offset: int = 0,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        file_format: str | None = "scarr",
        file_path: str | None = "auto",
    ):
        """
        Start run mode in the foreground, which will cause blocking.

        :param count: The number of traces to acquire.
        :type count: int
        :param sample_length: The sample length to be acquired.
        :type sample_length: int
        :param sample_offset: The sample offset to be acquired.
        :type sample_offset: int
        :param trigger_judge_wait_time: The trigger judge wait time in seconds.
        :type trigger_judge_wait_time: float
        :param trigger_judge_timeout: The trigger judge timeout in seconds.
        :type trigger_judge_timeout: float
        :param do_error_max_count: The maximum number of error count.
                                   If `do_error_handler_strategy` is 0, this setting is invalid.
        :type do_error_max_count: int
        :param do_error_handler_strategy: The strategy to handle error handling.
                                          0: Exit immediately, 1: Exit after exceeding the error count.
        :type do_error_handler_strategy: int
        :param file_format: The file format of the trace dataset. "scarr" or "numpy".
        :type file_format: str
        :param file_path: The file path of the trace dataset. If set to "auto", a folder with a timestamp format
                          will be created in the current working directory to save the data.
        :type file_path: str
        """
        try:
            self._do_run(
                test=False,
                count=count,
                sample_length=sample_length,
                sample_offset=sample_offset,
                trigger_judge_wait_time=trigger_judge_wait_time,
                trigger_judge_timeout=trigger_judge_timeout,
                do_error_max_count=do_error_max_count,
                do_error_handler_strategy=do_error_handler_strategy,
                file_format=file_format,
                file_path=file_path,
            )
            self._logger.debug("stop by timeout.")
        except KeyboardInterrupt:
            self._logger.debug("stop by interrupted.")
            self.stop()

    def test(
        self,
        count=-1,
        sample_length: int | None = None,
        sample_offset: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
    ):
        """
        Start test mode in background.

        :param count: The number of traces to acquire. normally this value should be -1.
        :type count: int
        :param sample_length: The sample length to be acquired.
        :type sample_length: int
        :param sample_offset: The sample offset to be acquired.
        :type sample_offset: int
        :param trigger_judge_wait_time: The trigger judge wait time in seconds.
        :type trigger_judge_wait_time: float
        :param trigger_judge_timeout: The trigger judge timeout in seconds.
        :type trigger_judge_timeout: float
        :param do_error_max_count: The maximum number of error count.
                                   If `do_error_handler_strategy` is 0, this setting is invalid.
        :type do_error_max_count: int
        :param do_error_handler_strategy: The strategy to handle error handling.
                                          0: Exit immediately, 1: Exit after exceeding the error count.
        """
        if self._status < 0:
            self._logger.debug("Resume to test mode.")
            self.resume()
        else:
            self._logger.debug("Start test mode.")
            self._run(
                test=True,
                count=count,
                sample_length=sample_length,
                sample_offset=sample_offset,
                trigger_judge_wait_time=trigger_judge_wait_time,
                trigger_judge_timeout=trigger_judge_timeout,
                do_error_max_count=do_error_max_count,
                do_error_handler_strategy=do_error_handler_strategy,
            )

    def test_sync(
        self,
        count=-1,
        sample_length: int | None = None,
        sample_offset: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
    ):
        """
        Start test mode in foreground, which will cause blocking.

        :param count: The number of traces to acquire. normally this value should be -1.
        :type count: int
        :param sample_length: The sample length to be acquired.
        :type sample_length: int
        :param sample_offset: The sample offset to be acquired.
        :type sample_offset: int
        :param trigger_judge_wait_time: The trigger judge wait time in seconds.
        :type trigger_judge_wait_time: float
        :param trigger_judge_timeout: The trigger judge timeout in seconds.
        :type trigger_judge_timeout: float
        :param do_error_max_count: The maximum number of error count.
                                   If `do_error_handler_strategy` is 0, this setting is invalid.
        :type do_error_max_count: int
        :param do_error_handler_strategy: The strategy to handle error handling.
                                          0: Exit immediately, 1: Exit after exceeding the error count.
        """
        try:
            self._do_run(
                test=True,
                count=count,
                sample_length=sample_length,
                sample_offset=sample_offset,
                trigger_judge_wait_time=trigger_judge_wait_time,
                trigger_judge_timeout=trigger_judge_timeout,
                do_error_max_count=do_error_max_count,
                do_error_handler_strategy=do_error_handler_strategy,
            )
        except KeyboardInterrupt:
            self._logger.debug("stop by interrupted.")
            self.stop()

    def pause(self):
        """
        Pause test or run mode.

        """
        if self._status > 0:
            self._run_thread_pause_event.clear()
            self._status = self._status * self._STATUS_PAUSED_SWITCH

    def resume(self):
        """
        Resume from paused test or run mode.

        """
        if self._status < 0:
            self._run_thread_pause_event.set()
            self._status = self._status * self._STATUS_PAUSED_SWITCH

    def stop(self):
        """
        Stop test or run mode.

        """
        if not self._run_thread_pause_event.is_set():
            self._run_thread_pause_event.set()
        self._status = self.STATUS_STOPPED

    def _run(
        self,
        test: bool = True,
        count: int = 1,
        sample_length: int | None = None,
        sample_offset: int | None = None,
        data_length: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        file_format: str | None = "scarr",
        file_path: str | None = "auto",
    ):
        self._run_thread_pause_event.set()
        threading.Thread(
            target=self._do_run,
            kwargs={
                "test": test,
                "count": count,
                "sample_length": sample_length,
                "sample_offset": sample_offset,
                "data_length": data_length,
                "trigger_judge_wait_time": trigger_judge_wait_time,
                "trigger_judge_timeout": trigger_judge_timeout,
                "do_error_max_count": do_error_max_count,
                "do_error_handler_strategy": do_error_handler_strategy,
                "file_format": file_format,
                "file_path": file_path,
            },
        ).start()

    def _do_run(
        self,
        test: bool = True,
        count: int = 1,
        sample_length: int | None = None,
        sample_offset: int | None = None,
        data_length: int | None = None,
        trigger_judge_wait_time: float | None = None,
        trigger_judge_timeout: float | None = None,
        do_error_max_count: int | None = None,
        do_error_handler_strategy: int | None = None,
        file_format: str | None = "scarr",
        file_path: str | None = "auto",
    ):
        if self._status > 0:
            self._logger.warning(
                f'AcquisitionTemplate is already running in {'run' if self._status == 2 else 'test'} mode.'
            )
            return

        if not test:
            self._status = self.STATUS_RUNNING
            self._status_changed()
        else:
            self._status = self.STATUS_TESTING
            self._status_changed()

        self.trace_count = count
        if sample_length is not None:
            self.sample_length = sample_length
        if sample_offset is not None:
            self.sample_offset = sample_offset
        if data_length is not None:
            self.data_length = data_length
        if trigger_judge_wait_time is not None:
            self.trigger_judge_wait_time = trigger_judge_wait_time
        if trigger_judge_timeout is not None:
            self.trigger_judge_timeout = trigger_judge_timeout
        if do_error_max_count is not None:
            self.do_error_max_count = do_error_max_count
        if do_error_handler_strategy is not None:
            self.do_error_handler_strategy = do_error_handler_strategy
        if file_format is not None:
            self.file_format = file_format
        if file_path is not None:
            self.file_path = file_path
        self.pre_init()
        self.init()
        self._post_init()
        self._loop(not test, self.file_path, self.file_format)
        self._pre_finish()
        self._finish()
        self._post_finish()

    def _status_changed(self):
        for listener in self._on_status_change_listeners:
            listener(self._status)

    def _loop(self, persistent: bool = True, file_path: str | None = None, file_format: str = "scarr"):
        do_error_count = 0
        trace_index = 0
        self._progress_changed(AcqProgress(trace_index, self.trace_count))

        cracker_version = self.cracker.get_firmware_version()
        if persistent:
            channel_names = [
                str(k) for k, v in self.cracker.get_current_config().osc_analog_channel_enable.items() if v
            ]
            if self.sample_length == -1:
                sample_length = self.cracker.get_current_config().osc_sample_length
            else:
                sample_length = self.sample_length

            if file_path is None or file_path == "auto":
                file_path = self._DATASET_DEFAULT_PATH
            if not file_path.endswith("/"):
                file_path += "/"
            file_path += datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            if file_format == "scarr":
                file_path += ".zarr"
            elif file_format == "numpy":
                file_path += ".npy"

            if file_format == "scarr":
                dataset = ScarrTraceDataset.new(
                    file_path,
                    channel_names,
                    self.trace_count,
                    sample_length,
                    self.data_length,
                    version=f"({cracker_version}, {cracker_version})",
                )  # todo channel
            elif file_format == "numpy":
                dataset = NumpyTraceDataset.new(
                    file_path,
                    channel_names,
                    self.trace_count,
                    sample_length,
                    self.data_length,
                    version=f"({cracker_version}, {cracker_version})",
                )  # todo channel
            else:
                self._logger.error(f"Unsupported file format: {file_format}")
                return
        else:
            dataset = None

        while self._status != 0 and self.trace_count - trace_index != 0:
            if self._status < 0:
                self._status_changed()
                self._run_thread_pause_event.wait()
                self._status_changed()
            self._logger.debug("Get wave data: %s", trace_index)
            self.pre_do()
            start = time.time()
            try:
                data = self.do()
            except Exception as e:
                self._logger.error(f"Do error <{e}>:\n {traceback.format_exc()}")
                do_error_count += 1
                if self.do_error_handler_strategy == self.DO_ERROR_HANDLER_STRATEGY_EXIT:
                    self._logger.error("Exit with do error")
                    break
                elif do_error_count > self.do_error_max_count:
                    self._logger.error(
                        f"Exit with do function get error count: {do_error_count} is grate than do_error_max_count"
                    )
                    break
                else:
                    self._logger.error(f"Do function get error count: {do_error_count}")
                    continue
            self._logger.debug(f"count: {trace_index} delay: {time.time() - start}")
            trigger_judge_start_time = time.time()
            while self._status != 0:
                trigger_judge_time = time.time() - trigger_judge_start_time
                if trigger_judge_time > self.trigger_judge_timeout:
                    self._logger.error(
                        "Triggered judge timeout and will get next waves, " "judge time: %s and timeout is %s",
                        trigger_judge_time,
                        self.trigger_judge_timeout,
                    )
                    break
                self._logger.debug("Judge trigger status.")
                if self._is_triggered():
                    self._last_wave = self._get_waves(self.sample_offset, self.sample_length)
                    if self._last_wave is not None:
                        self._logger.debug(
                            "Got wave: %s.",
                            {k: v.shape for k, v in self._last_wave.items()},
                        )
                        # self._logger.debug('Got wave: ')
                    if self._on_wave_loaded_callback and callable(self._on_wave_loaded_callback):
                        try:
                            self._on_wave_loaded_callback(self._last_wave)
                        except Exception as e:
                            self._logger.error("Wave loaded event callback error: %s", e.args)
                    break

                time.sleep(self.trigger_judge_wait_time)
            if dataset is not None and self._last_wave is not None:
                for k in self._last_wave.keys():
                    if data is not None and isinstance(data, bytes):
                        data = np.frombuffer(data, dtype=np.uint8)
                    dataset.set_trace(str(k), trace_index, self._last_wave[k], data)
            self._post_do()
            trace_index += 1
            self._current_trace_count = trace_index
            self._progress_changed(AcqProgress(trace_index, self.trace_count))
            # Reduce the execution frequency in test mode.
            if not persistent:
                time.sleep(2)

        if dataset is not None:
            dataset.dump()
        self._status = self.STATUS_STOPPED
        self._status_changed()

    def _progress_changed(self, progress: "AcqProgress"):
        for listener in self._on_run_progress_changed_listeners:
            listener({"finished": progress.finished, "total": progress.total})

    def _loop_without_log(self):
        ...
        # for i in range(self.trace_count):
        #     # print('yes', self._logger.handlers[0].stream)
        #     # self._logger.debug('Get wave data: %s', i)
        #     self.pre_do()
        #     self.do()
        #     trigger_judge_start_time = time.time()
        #     while self._status:
        #         trigger_judge_time = time.time() - trigger_judge_start_time
        #         if trigger_judge_time > self.trigger_judge_timeout:
        #             # self._logger.error("Triggered judge timeout and will get next waves, "
        #             #                    "judge time: %s and timeout is %s",
        #             #                    trigger_judge_time, self._trigger_judge_timeout)
        #             break
        #         # self._logger.debug('Judge trigger status.')
        #         if self._is_triggered():
        #             self._last_wave = self._get_waves()
        #             if self._last_wave is not None:
        #                 ...
        #                 # self._logger.debug('Got wave: %s.', self._last_wave.shape)
        #             if self._on_wave_loaded_callback and callable(self._on_wave_loaded_callback):
        #                 try:
        #                     self._on_wave_loaded_callback(self._last_wave)
        #                 except Exception as e:
        #                     ...
        #                     # self._logger.error('Wave loaded event callback error: %s', e.args)
        #             time.sleep(1)
        #             break
        #         time.sleep(self.trigger_judge_wait_time)
        #     self._save_wave()
        #     self._post_do()

    def connect_scrat(self):
        """
        Connect to scrat device
        :return:
        """
        ...

    def connect_cracker(self):
        """
        Connect to cracker device.
        :return:
        """
        self.cracker.connect()

    def connect_net(self): ...

    def config_scrat(self): ...

    def config_cracker(self): ...

    def pre_init(self): ...

    @abc.abstractmethod
    def init(self):
        """
        The `init` logic, which the user should implement in the subclass.

        """
        ...

    def _post_init(self): ...

    def transfer(self): ...

    def pre_do(self):
        self.cracker.osc_single()

    @abc.abstractmethod
    def do(self):
        """
        The `do` logic, which the user should implement in the subclass.

        """
        ...

    def _do(self):
        try:
            self.do()
            return True
        except Exception as e:
            self._logger.error("Do error: %s", e.args)
            return False

    def _post_do(self): ...

    def _is_triggered(self):
        _, triggered = self.cracker.osc_is_triggered()
        return triggered

    def _get_waves(self, offset: int, sample_length: int) -> dict[int, np.ndarray]:
        if sample_length == -1:
            sample_length = self.cracker.get_current_config().osc_sample_length
        config = self.cracker.get_current_config()
        if config.osc_analog_channel_enable is None:
            raise Exception("Channel info can't be none.")
        enable_channels = [k for k, v in config.osc_analog_channel_enable.items() if v]
        wave_dict = {}
        for c in enable_channels:
            status, wave_dict[c] = self.cracker.osc_get_analog_wave(c, offset, sample_length)
        return wave_dict

    def _pre_finish(self): ...

    def _post_finish(self): ...

    def _save_dataset(self): ...

    def _finish(self):
        pass

    def get_last_wave(self) -> dict[int, np.ndarray] | None:
        """
        Get the last wave.

        """
        return self._last_wave

    def dump_config(self, path: str = None) -> str | None:
        """
        Dump the current config to a JSON file if a path is specified, or to a JSON string if no path is specified.

        :param path: the path to the JSON file
        :return: the content of JSON string or None if no path is specified.
        """
        config = {
            "trace_count": self.trace_count,
            "sample_length": self.sample_length,
            "sample_offset": self.sample_offset,
            "trigger_judge_timeout": self.trigger_judge_timeout,
            "trigger_judge_wait_time": self.trigger_judge_wait_time,
            "do_error_max_count": self.do_error_max_count,
            "trace_file_path": self.file_path,
        }
        if path is None:
            return json.dumps(config)
        else:
            with open(path, "w") as f:
                json.dump(config, f)

    def load_config_from_file(self, path: str) -> None:
        """
        Load config from a JSON file.

        :param path: the path to the JSON file
        """
        with open(path) as f:
            self.load_config_from_str(json.load(f))

    def load_config_from_str(self, json_str: str) -> None:
        """
        Load config from a JSON string.

        :param json_str: the JSON string
        """
        config = json.loads(json_str)
        self.trace_count = config["trace_count"]
        self.sample_length = config["sample_length"]
        self.sample_offset = config["sample_offset"]
        self.trigger_judge_timeout = config["trigger_judge_timeout"]
        self.trigger_judge_wait_time = config["trigger_judge_wait_time"]
        self.do_error_max_count = config["do_error_max_count"]
        self.file_path = config["trace_file_path"]

    def is_running(self):
        return self._status != self.STATUS_STOPPED

    @staticmethod
    def builder():
        return AcquisitionBuilder()


class AcquisitionBuilder:
    """
    The helper class for create a Sub acquisition class.

    """

    def __init__(self):
        self._cracker = None
        self._do_function = lambda _: ...
        self._init_function = lambda _: ...

    def cracker(self, cracker: CrackerBasic):
        """
        Specify the cracker for acquisition.

        :param cracker: the cracker
        :type cracker: CrackerBasic
        :return: 'AcquisitionBuilder'
        """
        self._cracker = cracker
        return self

    def init(self, init_function: typing.Callable[[CrackerBasic], None]):
        """
        The init function of acquisition.

        :param init_function: the init function
        :type init_function: typing.Callable[[CrackerBasic], None]
        :return: 'AcquisitionBuilder'
        """
        if init_function is not None:
            self._init_function = init_function
        return self

    def do(self, do_function: typing.Callable[[CrackerBasic], None]):
        """
        The do function of acquisition.

        :param do_function: the do function
        :type do_function: typing.Callable[[CrackerBasic], None]
        :return: 'AcquisitionBuilder'
        """
        if do_function is not None:
            self._do_function = do_function
        return self

    def build(self, **kwargs):
        """
        Build the acquisition object.

        :param kwargs: other keyword arguments of Acquisition.
        :type kwargs: dict
        :return Acquisition
        """
        if self._cracker is None:
            raise ValueError("No cracker")

        builder_self = self

        class AnonymousAcquisition(Acquisition):
            def __init__(self, *ana_args, **ana_kwargs):
                super().__init__(*ana_args, **ana_kwargs)

            def init(self):
                builder_self._init_function(self.cracker)

            def do(self):
                return builder_self._do_function(self.cracker)

        return AnonymousAcquisition(self._cracker, **kwargs)



# Copyright 2024 CrackNuts. All rights reserved.

import logging
import socket
import struct
import sys
import threading

from cracknuts import logger
from cracknuts.cracker import protocol, operator
from cracknuts.utils import hex_util


_handler_dict = {}


def _handler(command: int, has_payload: bool = True):
    def decorator(func):
        def wrapper(*args, **kwargs):
            if not has_payload:
                del kwargs["payload"]
            return func(*args, **kwargs)

        _handler_dict[command] = wrapper
        return wrapper

    return decorator


class MockOperator:
    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = protocol.DEFAULT_OPERATOR_PORT,
        logging_level=logging.INFO,
        init_status: int = 0,
    ):
        self._running = False
        self._status = init_status.to_bytes()
        self._logger = logger.get_logger(self, logging_level)
        self._server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server_socket.settimeout(1)
        self._server_socket.bind((host, port))
        self._server_socket.listen()
        self._logger.debug("MockCracker initialized")

    def start(self):
        try:
            self._running = True
            self._start_accept()
        except KeyboardInterrupt:
            self._logger.debug("exist by keyboard interrupt...")
            sys.exit(0)

    def stop(self):
        self._running = False

    def _start_accept(self):
        while self._running:
            try:
                conn, addr = self._server_socket.accept()
                conn.settimeout(5)
                self._logger.debug(f"Get client request from {addr}.")
                threading.Thread(target=self._handle, args=(conn,), daemon=True).start()
                # self._handle(conn)
            except TimeoutError:
                continue

    def _handle(self, conn):
        while self._running:
            try:
                header_data = conn.recv(protocol.REQ_HEADER_SIZE)
                if not header_data:
                    self._logger.debug("disconnected...")
                    break

                self._logger.debug(f"Received header:\n{hex_util.get_bytes_matrix(header_data)}")

                try:
                    magic, version, direction, command, rfu, length = struct.unpack(
                        protocol.REQ_HEADER_FORMAT, header_data
                    )
                except struct.error as e:
                    err_msg = f"Message format error: {e.args[0]}\n"
                    self._logger.error(err_msg)
                    err_res = self._error(f"Header format error: {hex_util.get_hex(header_data)}")
                    self._logger.debug(f"Send error:\n{hex_util.get_bytes_matrix(err_res)}")
                    conn.sendall(err_res)
                    continue

                self._logger.info(f"Received command: 0x{command:02X}")
                self._logger.debug(
                    f"Received header: Magic={magic}, Version={version}, Direction={direction}, "
                    f"Command={command}, Length={length}"
                )

                payload_data = conn.recv(length)
                if len(payload_data) > 0:
                    self._logger.debug(f"Received payload:\n{hex_util.get_bytes_matrix(payload_data)}")
                if command not in _handler_dict:
                    self._logger.warning(f'Get command not supported: 0x{format(command, '04x')}')
                    unsupported_res = self._unsupported(command)
                    self._logger.debug(f"Send unsupported:\n{hex_util.get_bytes_matrix(unsupported_res)}")
                    conn.sendall(unsupported_res)
                    continue
                func_res = _handler_dict.get(command, None)(self, payload=payload_data)
                if type(func_res) is tuple:
                    status, res_payload = func_res
                else:
                    status, res_payload = protocol.STATUS_OK, func_res
                res = protocol.build_response_message(status, res_payload)
                self._logger.debug(f"Sending:\n{hex_util.get_bytes_matrix(res)}")
                conn.sendall(res)
            except TimeoutError:
                continue
            except ConnectionResetError:
                break
            except Exception as e:
                self._logger.error(e)
                raise e

    @staticmethod
    def _error(error: str) -> bytes:
        return protocol.build_response_message(protocol.STATUS_ERROR, error.encode())

    @staticmethod
    def _unsupported(command):
        return protocol.build_response_message(
            protocol.STATUS_COMMAND_UNSUPPORTED, f'Command [0x{format(command, '04x')}] not supported'.encode()
        )

    @_handler(operator.Command.GET_STATUS, has_payload=False)
    def get_status(self) -> bytes:
        return self._status

    @_handler(operator.Command.START_SERVER, has_payload=False)
    def start_server(self) -> bytes:
        return b""

    @_handler(operator.Command.STOP_SERVER, has_payload=False)
    def stop_server(self) -> bytes:
        return b""

    @_handler(operator.Command.UPDATE_SERVER)
    def update_server(self, payload: bytes) -> bytes:
        self._status = b"\x01"
        return b""

    @_handler(operator.Command.UPDATE_BITSTREAM)
    def update_bitstream(self, payload: bytes) -> bytes:
        return b""

    @_handler(operator.Command.GET_MODEL, has_payload=False)
    def get_model(self):
        return b"mock_s1"

    @_handler(operator.Command.GET_SN, has_payload=False)
    def get_sn(self):
        return b"0001"

    @_handler(operator.Command.GET_SERVER_VERSION, has_payload=False)
    def get_server_version(self):
        return b"mock server"

    @_handler(operator.Command.GET_VERSION, has_payload=False)
    def get_version(self):
        return b"mock operator"



# Copyright 2024 CrackNuts. All rights reserved.

from enum import Enum


class Stopbits(Enum):
    STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO = 0, 1, 2


class Parity(Enum):
    PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE = 0, 1, 2, 3, 4


class Bytesize(Enum):
    FIVEBITS, SIXBITS, SEVENBITS, EIGHTBITS = 5, 6, 7, 8


class Baudrate(Enum):
    BAUDRATE_9600, BAUDRATE_19200, BAUDRATE_38400, BAUDRATE_57600, BAUDRATE_115200, BAUDRATE_1000000 = 0, 1, 2, 3, 4, 5


class SpiCpol(Enum):
    SPI_CPOL_LOW, SPI_CPOL_HIGH = 0, 1


class SpiCpha(Enum):
    SPI_CPHA_LOW, SPI_CPHA_HIGH = 0, 1


class I2cSpeed(Enum):
    STANDARD_100K, FAST_400K, FAST_PLUS_1M, HIGH_SPEED_3400K, ULTRA_FAST_5M = 0, 1, 2, 3, 4
